/**
  *  \file arch/generic/file.cc
  *  \brief Generic implementation of the file stream class
  *
  *  This uses C++'s filestreams to implement file access. If
  *  possible, a platform-specific version should be used: we usually
  *  don't need buffering. Actually, we often add an additional level
  *  of buffering atop the stream.
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <fstream>
#include <cstring>
#include <cerrno>
#include "io/streams.h"
#include "util/except.h"

namespace {
    class GenericFileStream : public Stream {
        std::fstream file;
        string_t name;
        void init(const char* aname, int amode);
     public:
        GenericFileStream(string_t aname, int amode);
        ~GenericFileStream();
    
        std::size_t read(char* buffer, std::size_t max);
        std::size_t write(const char* buffer, std::size_t max);
        int getCapabilities() const;
        void seek(fpos_t pos);
        fpos_t getPos();
        fpos_t getSize();
        string_t getName();
    };
}

Ptr<Stream>
Stream::openFile(const string_t& name, int mode)
{
    return new GenericFileStream(name, mode);
}

/** \class GenericFileStream
    \brief Stream reading from a file
    
    This is a file stream implemented atop a standard fstream. */
GenericFileStream::GenericFileStream(string_t file_name, int mode)
    : name(file_name)
{
    init(file_name.c_str(), mode);
}

GenericFileStream::~GenericFileStream()
{
}

void
GenericFileStream::init(const char* file_name, int mode)
{
    int open_mode = 0;
    if(mode & C_READ)
        open_mode |= std::ios::in;
    if(mode & C_WRITE)
        open_mode |= std::ios::out;

    // We do our own text file parsing
    open_mode |= std::ios::binary;

    if(mode & C_CREATE)
        open_mode |= (std::ios::trunc | std::ios::in | std::ios::out);
    else
        open_mode |= std::ios::in;

    errno = 0;
    file.open(file_name, std::ios::openmode(open_mode));
    if(!file) {
        throw FileOpenException(file_name, strerror(errno));
    }
}

std::size_t
GenericFileStream::read(char* buffer, std::size_t max)
{
    file.clear();
    if(file.read(buffer, max))  // FIXME: read error?
        return max;
    else
        return file.gcount();
}

std::size_t
GenericFileStream::write(const char* buffer, std::size_t max)
{
    file.clear();
    if(file.write(buffer, max)) // FIXME: write error?
        return max;
    else
        return 0;
}

void
GenericFileStream::seek(fpos_t pos)
{
    /* CD2:27.8.1.1#3: fstreams have only one file position */
    file.clear();
    file.seekg(pos);
    if (!file.good()) {
        notSeekable();
    }
}

GenericFileStream::fpos_t
GenericFileStream::getPos()
{
    file.clear();
    return file.tellg();
}

GenericFileStream::fpos_t
GenericFileStream::getSize()
{
    file.clear();
    fpos_t p = file.tellg();
    fpos_t res = -1;
    if(file.seekg(0, std::ios::end))
        res = file.tellg();
    file.seekg(p);
    file.clear();
    return res;
}

int
GenericFileStream::getCapabilities() const
{
    return C_READ | C_WRITE | C_SEEK; // FIXME?
}

string_t
GenericFileStream::getName()
{
    return name;
}
