/**
  *  \file client/dialogs/exportparam.cc
  *  \brief Export Parameter dialog
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/exportparam.h"
#include "client/widgets/expfieldlist.h"
#include "client/widgets/expformat.h"
#include "client/widgets/helpwidget.h"
#include "cpluslib/format.h"
#include "gfx/palette.h"
#include "int/export.h"
#include "int/interror.h"
#include "io/glob.h"
#include "ui/dialogs/filedlg.h"
#include "ui/layout.h"
#include "ui/msgbox.h"
#include "ui/uiglobal.h"
#include "ui/widgets/button.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/statictext.h"
#include "util/nls.h"
#include "ui/widgets/quit.h"

/************************* WExportParameterDialog ************************/

WExportParameterDialog::WExportParameterDialog(IntExportFieldList& fields,
                                               IntContext* ctx,
                                               AnswerProvider& filter)
    : UIWindow(_("Export"), BLUE_WINDOW, COLOR_GRAY),
      fields(fields),
      ctx(ctx),
      filter(filter)
{
    init();
}

void
WExportParameterDialog::init()
{
    // VBox
    //   UIFrameGroup/HBox
    //     WExportFieldList
    //     UIScrollbar
    //   HBox
    //     UIButton "Ins"
    //     UIButton "Del"
    //     UISpacer
    //  [fileformatselect]
    //   HBox
    //     "OK"
    //     "Cancel"
    //     UISpacer
    //     "Help"

    setLayout(UIVBoxLayout::instance5);

    UIFrameGroup& g1 = h.add(new UIFrameGroup(UIHBoxLayout::instance0));
    UIGroup&      g2 = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup&      g4 = h.add(new UIGroup(UIHBoxLayout::instance5));

    // Field list
    WExportFieldList& xfl = h.add(new WExportFieldList(fields, ctx));
    g1.add(xfl);
    g1.add(h.add(new UIScrollbar(xfl)));
    add(g1);

    // Field list buttons
    UIButton* btn = &h.add(new UIButton(_("Ins"), SDLK_INSERT));
    btn->setState(UIButton::bf_SmallSize, true);
    g2.add(*btn);
    g2.add(h.add(new UIStaticText(_("Add..."), UIStaticText::StaticNormal)));

    btn = &h.add(new UIButton(_("Del"), SDLK_INSERT));
    btn->setState(UIButton::bf_SmallSize, true);
    g2.add(*btn);
    g2.add(h.add(new UIStaticText(_("Remove"), UIStaticText::StaticNormal)));

    g2.add(h.add(new UISpacer()));

    g2.add(h.add(new UIStaticText(_("Width"), UIStaticText::StaticNormal)));
    btn = &h.add(new UIButton("+", '+'));
    btn->setState(UIButton::bf_SmallSize, true);
    g2.add(*btn);
    btn = &h.add(new UIButton("-", '-'));
    btn->setState(UIButton::bf_SmallSize, true);
    g2.add(*btn);

    add(g2);

    // File format select
    format_control = &h.add(new WExportFormatControl());
    add(*format_control);

    // Dialog buttons
    g4.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addCall(this, &WExportParameterDialog::onOK));
    g4.add(h.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    g4.add(h.add(new UISpacer()));
    g4.add(h.add(new UIButton(_("Help"), 'h')));
    add(g4);

    add(h.add(new WHelpWidget("pcc2:export")));
    add(h.add(new UIQuit(0)));

    pack();
}

/** OK button. */
void
WExportParameterDialog::onOK()
{
    // Must have some fields
    if (fields.size() == 0) {
        messageBox(_("Please select some fields for output."), _("Export"));
        return;
    }

    // Select output file name
    string_t ext = format_control->getRecommendedFileExtension();
    UIFileChooserDialog dlg(_("Export"),
                            Directory::openDir(Directory::getWorkingDirectory()),
                            FileNamePattern::getAllFilesWithExtensionPattern(ext));
    ui_root->centerWidget(dlg);
    if (ui_root->runModal(dlg)) {
        string_t name = appendFileNameExtension(dlg.getResultFile(), ext, false);
        try {
            Ptr<Stream> s = Stream::openFile(name, Stream::C_CREATE);
            format_control->doExport(*s, ctx, filter, fields);
            stop(1);
        }
        catch (IntError& e) {
            messageBox(format(_("Error during export: %s"), e.what()),
                       _("Export"));
        }
        catch (GError& e) {
            messageBox(format(_("Error during export: %s"), e.getUserError()),
                       _("Export"));
        }
        catch (FileProblemException& e) {
            messageBox(format(_("Unable to create %s: %s"), e.getFileName(), e.what()),
                       _("Export"));
        }
        catch (std::runtime_error& e) {
            messageBox(format(_("Unable to create %s: %s"), name, e.what()),
                       _("Export"));
        }
    }
}

