/**
  *  \file client/history.cc
  *  \brief Screen History
  *
  *  This module maintains a history of the most recent screens seen
  *  by the user. This is used to implement the 'Backspace' and
  *  'Alt-Backspace' functions.
  *
  *  We implement this feature as a vector of bounded size. Since it's
  *  possible that entries get invalid, we garbage-collect those on
  *  every public API function. In case we switch to an unbounded list
  *  someday, we'd have to be smarter here. Right now it's fast enough
  *  that it doesn't matter.
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <vector>
#include "client/history.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/window.h"
#include "util/nls.h"
#include "util/modinit.h"
#include "int/if/userif.h"
#include "ui/widgets/button.h"
#include "ui/widgets/quit.h"

enum {
    /** Maximum size of screen history. */
    SCREEN_HISTORY_MAX_SIZE = 50
};

/** Screen history. */
static std::vector<WScreenId> screen_history;

/** Rotate history by one element. Turns situation A-B-C-D-E into
    E-A-B-C-D. */
static void
rotate()
{
    WScreenId last = screen_history.back();
    for (std::vector<WScreenId>::size_type i = screen_history.size(); i > 1; --i)
        screen_history[i-1] = screen_history[i-2];
    screen_history[0] = last;
}

/** Discard all bogus entries. \see WScreenId::isValid */
static void
discardBogusEntries()
{
    std::vector<WScreenId>::iterator i = screen_history.begin();
    while (i != screen_history.end()) {
        if (!i->isValid())
            i = screen_history.erase(i);
        else
            ++i;
    }
}

/** Clear screen history. User upon every entry/exit of the race
    screen. This makes sure we start with a blank list*/
void
clearScreenHistory()
{
    screen_history.clear();
    screen_history.reserve(SCREEN_HISTORY_MAX_SIZE);
}

/** Push new history Id. This expresses that we're now in the
    specified context. This function is idempotent in that it refuses
    to push duplicates, and also tries to keep the history concise in
    other ways. */
void
pushHistoryScreenId(WScreenId id)
{
    discardBogusEntries();

    /* If this is the same as what we already have on top, don't push anything */
    if (screen_history.size() > 0 && screen_history.back() == id)
        return;

    /* If this is a starchart entry, and the previous one is the starchart as
       well, replace it */
    if (id.getScreenId() == WScreenId::Starchart
        && screen_history.size() > 0 && screen_history.back().getScreenId() == WScreenId::Starchart)
    {
        screen_history.back() = id;
        return;
    }

    /* If we have a situation A-B-A, and this is B, pop an A instead. 
       This avoids that the history clutters up with As and Bs if the
       user rapidly switches between these two.
       FIXME: make this smarter to detect any-size cycles */
    if (screen_history.size() > 2 && screen_history[screen_history.size()-2] == id
        && screen_history[screen_history.size()-1] == screen_history[screen_history.size()-3])
    {
        screen_history.pop_back();
        return;
    }

    /* If this would overflow the size, drop the first item */
    if (screen_history.size() >= SCREEN_HISTORY_MAX_SIZE)
        screen_history.erase(screen_history.begin());
    screen_history.push_back(id);
}

/** Pop history Id. Invokes the current IntUserInterfaceBinding to perform
    actual screen change.
    \param current_screen_registered true iff current screen is registered on
    the history stack, false if not.
    \return true if we did anything */
bool
popHistoryScreenId(bool current_screen_registered)
{
    discardBogusEntries();

    /* Figure out target screen */
    WScreenId scr;
    if (screen_history.size() <= 0) {
        scr = WScreenId();
    } else if (screen_history.size() == 1) {
        scr = screen_history.front();
    } else {
        /* If the current screen is registered, it lies on top of the
           stack; rotate it to the bottom */
        if (current_screen_registered)
            rotate();
        scr = screen_history.back();
    }

    if (scr.isNonNull()) {
        if (scr.getScreenId() == WScreenId::Starchart) {
            postGoToChart(IntUserInterfaceBinding::get(), scr.getPos());
        } else {
            postGoToScreen(IntUserInterfaceBinding::get(), scr.getScreenId(), scr.getId());
        }
        return true;
    } else {
        return false;
    }
}

namespace {
    class WScreenHistoryList : public UIListbox {
        bool current_screen_registered;
     public:
        WScreenHistoryList(bool current_screen_registered);
        void drawPart(GfxCanvas& can, int from, int to);
    };
}

WScreenHistoryList::WScreenHistoryList(bool current_screen_registered)
    : UIListbox(0, 20, 300, screen_history.size()-current_screen_registered),
      current_screen_registered(current_screen_registered)
{ }

void
WScreenHistoryList::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);
    for (int i = from; i < to; ++i) {
        /* we display the list backwards, and line 0 in the list box is
           screen_history.size()-2 */
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);
        int index = (screen_history.size()-1-current_screen_registered) - (i + getTop());
        if (index >= 0) {
            outTextF(ctx, area.x, area.y, area.w, screen_history[index].toString());
        }
    }
}



bool
doScreenHistoryDialog(bool current_screen_registered)
{
    discardBogusEntries();

    /* trivial case */
    if (screen_history.size() <= 1)
        return false;

    /* build list and standard dialog */
    WScreenHistoryList list(current_screen_registered);
    ObjectHolder h;
    UIWindow win(_("Screen History"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
    win.setLayout(UIVBoxLayout::instance5);
    win.add(list);

    UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
    g.add(h.add(new UISpacer()));
    g.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    g.add(h.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    win.add(g);
    win.add(h.add(new UIQuit(0)));

    win.pack();
    ui_root->centerWidget(win);
    ui_root->add(win);
    bool rv = ui_root->run();
    ui_root->close(win);

    if (rv) {
        /* this is nominally O(n^2), but our n is small, so keep it simple */
        int n = list.getCurrent();
        while (n-- > 0)
            rotate();
        return popHistoryScreenId(current_screen_registered);
    } else {
        return false;
    }
}

extern const ModuleInitializer mod_history[] = {
    { OnEnterDirectory | OnEnterGame | OnLeaveGame, clearScreenHistory },
    { 0, 0 },
};
