/**
  *  \file client/tiles/shipmission.cc
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <memory>
#include "client/tiles/shipmission.h"
#include "ui/uiglobal.h"
#include "game/mission.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "game/ship.h"
#include "game/planet.h"
#include "game/config.h"
#include "game/univ.h"
#include "int/intglobal.h"
#include "int/value.h"
#include "int/data.h"
#include "game/playerset.h"
#include "game/racename.h"
#include "ui/table.h"
#include "game/formula.h"
#include "game/host.h"
#include "game/tables.h"
#include "ui/widgets/button.h"
#include "game/shippredict.h"
#include "util/misc.h"

/**************************** WShipMissionTile ***************************/

WShipMissionTile::WShipMissionTile(GObjectSelection& selection)
    : WObjectSelectionObserverTile(GfxPoint(315, 3*font_heights[FONT_NORMAL]), selection)
{
    setLayout(RightJustified, 0, false);
    addControl(GfxRect(27, 0, 20, 20),  h.add(new UIButton("M", 'm')));
    addControl(GfxRect(27, 22, 20, 20), h.add(new UIButton("E", 'e')));
    addControl(GfxRect(27, 44, 20, 20), h.add(new UIButton("F", 'f')));

    cf.add(h.add(new UIButton(cm_Ship_BaseMenu, "B", 'b')).addCommand());
    addControl(GfxRect(0, 42, 24, 24), cf);
}

void
WShipMissionTile::drawData(GfxCanvas& can)
{
    const int y = font_heights[FONT_NORMAL];
    const int x = 120;

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.useColorScheme(getSkin());
    ctx.setSolidBackground();
    ctx.setColor(UIColor::tc_Static);

    outText(ctx, 5,   0, _("Mission:"));
    outText(ctx, 5,   y, _("Primary Enemy:"));
    outText(ctx, 5, 2*y, _("Friendly Code:"));

    // cannot be const.
    GShip* s = dynamic_cast<GShip*>(getCurrentObject());
    if (!s)
        return;
    GShip& ship = *s;

    if (ship.getMission().isKnown()) {
        const GMission* msn = GMissionList::getInstance().getMissionByShip(ship);
        ctx.setColor(UIColor::tc_Green);
        if (msn) {
            if (msn->isWarning(ship)) {
                ctx.setColor(UIColor::tc_Yellow);
            }
            outTextF(ctx, x-50, 0, 214, msn->getLabel(ship));
        } else {
            outTextF(ctx, x-50, 0, 214, format("M%d I%d T%d", ship.getMission(), ship.getInterceptId(), ship.getTowId()));
        }
    }

    ctx.setColor(UIColor::tc_Green);
    if (ship.getPrimaryEnemy().isKnown()) {
        outTextF(ctx, x,    y, 164-25,
                 (ship.getPrimaryEnemy()
                  ? player_racenames.getShortName(ship.getPrimaryEnemy())
                  : string_t(_("none"))));
    }

    if (ship.isFCodeKnown()) {
        outTextF(ctx, x,  2*y, 100, ship.getFCode());
    }

    // Set color of the 'b' button
    TFrameColor theColor = cf_None;
    if (GUniverse* univ = getCurrentUniverse()) {
        int baseId = ship.getOrbitPlanetId();
        if (baseId != 0 && univ->isValidPlanetId(baseId)) {
            GPlanet& planet = univ->getPlanet(baseId);
            if (planet.isPlayable(GObject::Playable) && planet.hasBase()) {
                // Planet has a base
                if (planet.getBaseShipyardId() == ship.getId()) {
                    if (planet.getBaseShipyardOrder() == sya_Recycle)
                        theColor = cf_Red;
                    else if (planet.getBaseShipyardOrder() == sya_Fix)
                        theColor = cf_Green;
                }
                // FIXME: handle cloning. PCC 1.x uses green if cloning, yellow if conflict
                // (multiple cloners)
            }
        }
    }
    cf.setColor(theColor);
}

/************************* WNarrowShipMissionTile ************************/

WNarrowShipMissionTile::WNarrowShipMissionTile(GObjectSelection& selection)
    : WObjectSelectionObserverTile(GfxPoint(250, 6*font_heights[FONT_NORMAL]), selection)
{ }

void
WNarrowShipMissionTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    GShip* sh = dynamic_cast<GShip*>(getCurrentObject());
    if (!sh) {
        drawBackground(ctx, getInnerSize(0, 0));
        return;
    }

    UITableWriter tw(ctx, getInnerSize(0, 0), FONT_NORMAL);
    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Enemy:"));
    tw.writeFixedCell(UIColor::tc_Static, _("FCode:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Mission:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Course:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Speed:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Usage:"));

    tw.newLColumn();

    // PE
    if (sh->getPrimaryEnemy().isKnown()) {
        tw.writeVariableCell(UIColor::tc_Green,
                             (sh->getPrimaryEnemy()
                              ? player_racenames.getShortName(sh->getPrimaryEnemy()) : string_t(_("none"))));
    } else {
        tw.writeVariableCell(UIColor::tc_Green, string_t());
    }

    // FC
    if (sh->isFCodeKnown())
        tw.writeVariableCell(UIColor::tc_Green, sh->getFCode());
    else
        tw.writeVariableCell(UIColor::tc_Green, string_t());

    // Mission
    if (sh->getMission().isKnown()) {
        const GMission* msn = GMissionList::getInstance().getMissionByShip(*sh);
        ctx.setColor(UIColor::tc_Green);
        if (msn)
            tw.writeVariableCell(UIColor::tc_Green, msn->getLabel(*sh));
        else
            tw.writeVariableCell(UIColor::tc_Green,
                                 format("M%d I%d T%d", sh->getMission(), sh->getInterceptId(), sh->getTowId()));
    } else {
        tw.writeVariableCell(UIColor::tc_Green, string_t());
    }

    // Course
    if (sh->getShipKind() == GShip::CurrentShip) {
        double dist = distFromDX(sh->getWaypointDX(), sh->getWaypointDY());
        tw.writeVariableCell(!sh->isHyperdriving()
                             ? UIColor::tc_Green
                             : dist < 20
                             ? UIColor::tc_Red
                             : !host.isExactHyperjumpDistance2(int32(dist*dist+.1)) /* FIXME: this may lose precision, hence the +.1 */
                             ? UIColor::tc_Yellow
                             : UIColor::tc_Green,
                             format(_("%.2f ly"), dist));
    } else if (sh->getHeading().isKnown()) {
        tw.writeVariableCell(UIColor::tc_Green,
                             format(_("%d" "\xC2\xB0" ", %s"), sh->getHeading(), getHeadingName(sh->getHeading())));
    } else {
        tw.writeVariableCell(UIColor::tc_Green, string_t());
    }

    // Warp
    if (sh->getWarp().isKnown()) {
        if (sh->getWarp() == 0) {
            tw.writeVariableCell(sh->getShipKind() == GShip::CurrentShip && (sh->getWaypointDX() != 0 || sh->getWaypointDY() != 0) ? UIColor::tc_Red : UIColor::tc_Green,
                                 _("not moving"));
        } else if (sh->isHyperdriving()) {
            tw.writeVariableCell(UIColor::tc_Green, _("Hyperdrive"));
        } else {
            tw.writeVariableCell(UIColor::tc_Green, format(_("Warp %d"), sh->getWarp()));
        }
    } else {
        tw.writeVariableCell(UIColor::tc_Green, string_t());
    }

    // Fuel usage
    if (sh->getShipKind() == GShip::CurrentShip) {
        GUniverse& univ = *getCurrentUniverse();
        GShip& s = *sh;
        std::auto_ptr<GShipTurnPredictor> main, towee;
        if (s.getMission() == GMission::msn_Tow && univ.ty_played_ships.isValidIndex(s.getTowId())) {
            towee.reset(new GShipTurnPredictor(univ, s.getTowId()));
            main.reset(new GShipTurnPredictor(univ, s.getId(), *towee));
        } else {
            main.reset(new GShipTurnPredictor(univ, s.getId()));
        }
        main->computeMovement();

        if (sh->getWaypointDX() == 0 && sh->getWaypointDY() == 0)
            tw.writeVariableCell(UIColor::tc_Green, _("at waypoint"));
        else if (sh->isHyperdriving())
            tw.writeVariableCell(UIColor::tc_Green, _("Hyperdrive"));
        else if (sh->getWarp() == 0)
            tw.writeVariableCell(UIColor::tc_Red, _("not moving"));
        else if (main->isAtTurnLimit())
            tw.writeVariableCell(UIColor::tc_Yellow, _("too long"));
        else
            tw.writeVariableCell((main->getMovementFuelUsed() > sh->getCargo(el_Neutronium)) ? UIColor::tc_Red : UIColor::tc_Green,
                                 format(_("%d turn%!1{s%}, %d kt fuel"),
                                        numToString(main->getNumTurns()),
                                        numToString(main->getMovementFuelUsed())));
    } else {
        tw.writeVariableCell(UIColor::tc_Green, string_t());
    }
}
