/**
  *  \file client/widgets/fleetlist.cc
  *  \brief Fleet List
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "client/widgets/fleetlist.h"
#include "game/fleet.h"
#include "game/global.h"
#include "game/ship.h"
#include "gfx/gfxcontext.h"
#include "util/nls.h"

/******************************* WFleetList ******************************/

/** Constructor. */
WFleetList::WFleetList()
    : initial_selection(0)
{ }

WFleetList::~WFleetList()
{ }

/** Generate fleet list.
    \param univ   [in] Universe
    \param pt     [in] Location to check for
    \param flags  [in] Flags
    \param except [in] Do not list the fleet with this Id */
void
WFleetList::listFleets(const GUniverse& univ, const GPoint pt, const int flags, const int except)
{
    fleet_ids.clear();
    initial_selection = 0;

    for (uint16_t i = 1; i <= NUM_SHIPS; ++i) {
        const GShip& sh = univ.getShip(i);
        if (i != except && sh.isPlayable(GObject::Playable) && sh.isFleetLeader()) {
            bool here = (sh.getPos() == pt);
            if (here || (flags & fAllFleets) != 0) {
                if (initial_selection == 0 && here) {
                    initial_selection = fleet_ids.size();
                }
                uint16_t value = i;
                if (here) {
                    value |= mHere;
                }
                fleet_ids.push_back(value);
            }
        }
    }
}

/**************************** WFleetListWidget ***************************/

/** Fleet list widget.
    \param list Fleet list to display (must live as long as widget)
    \param nlines Number of lines to use on screen */
WFleetListWidget::WFleetListWidget(const WFleetList& list, const int nlines)
    : UIListbox(0, nlines, 30*font_ems[FONT_NORMAL], list.size()),
      list(list)
{
    scrollTo(list.getInitialSelection(), 1);
}

WFleetListWidget::~WFleetListWidget()
{ }

void
WFleetListWidget::drawPart(GfxCanvas& can, int from, int to)
{
    const char* hereMark = _("(here)");
    int totalWidth = getExtent().w;
    int hereWidth = std::min(totalWidth, fonts[FONT_NORMAL]->getTextWidth(hereMark) + 30);
    totalWidth -= hereWidth;
    int indentWidth = std::min(5, totalWidth);
    totalWidth -= indentWidth;
    
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);
    for (int i = from; i < to; ++i) {
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);

        std::size_t index = getTop() + i;
        if (index < list.size()) {
            outTextF(ctx, area.x + indentWidth, area.y, totalWidth, getFleetTitle(getDisplayedTurn().getCurrentUniverse(), list[index]));
            if (list.isHere(index)) {
                outTextF(ctx, area.x + indentWidth + totalWidth, area.y, hereWidth, hereMark);
            }
        }
    }
}


/****************************** doFleetList ******************************/

/** Fleet list standard dialog.
    \param title  [in] Dialog title
    \param univ   [in] Universe
    \param pt     [in] Location to check for
    \param flags  [in] Flags
    \param except [in] Do not list the fleet with this Id
    \return chosen fleet Id, 0 if none or canceled */
int
doFleetList(const string_t title, GUniverse& univ, GPoint pt, int flags, int except)
{
    // Generate list
    WFleetList list;
    list.listFleets(univ, pt, flags, except);

    // Quick exit
    if (list.size() == 0) {
        return 0;
    }
    if (list.size() == 1 && (flags & WFleetList::fQuickOut) != 0) {
        return list[0];
    }

    // Build it
    WFleetListWidget wid(list, list.size() < 5 ? 5 : list.size() > 15 ? 15 : list.size());
    if (wid.doStandardListbox(title, "pcc2:fleetscreen")) {
        return list[wid.getCurrent()];
    } else {
        return 0;
    }
}
