/**
  *  \file client/widgets/msgaction.cc
  *  \brief WMessageActionPanel widget
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include <cstring>
#include "client/widgets/msgaction.h"
#include "cpluslib/format.h"
#include "game/msgparse.h"
#include "game/player.h"
#include "game/racename.h"
#include "gfx/gfxcontext.h"
#include "gfx/palette.h"
#include "int/if/userif.h"
#include "resmgr/resmgr.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "util/nls.h"
#include "util/unicodechars.h"


/** Find X,Y coordinate pair.
    \param str [in] String
    \param n   [in] Index into string, position of a ','
    \param start,end [out] Position of result coordinate in string
    \param pt  [out] Parsed coordinate
    \return true if found, false if this is not a valid coordinate pair */
bool
findXY(const string_t& str, string_t::size_type n,
       string_t::size_type& start, string_t::size_type& end,
       GPoint& pt)
{
    /* Punctuation required before/after coordinate pair.
       The idea is to accept things like
          1111,1111
          (1111,1111)
          move to->1111,1111
          ship @1111,1111
       but reject things like
          map size 2000x2000, 50 ly seam
       (which would be interpreted as '2000,50' without the
       punctuation filter). Since this function gets a whole
       message, we must also accept '\n' as whitespace. */
    static const char PUNCT[] = "\n ():@-<=>.!";

    /* Locate X before n */
    int x = 0;
    int xdig = 0;
    int xmul = 1;
    string_t::size_type xn = n;
    while (xn > 0 && str[xn-1] == ' ') {
        /* Skip space */
        --xn;
    }
    while (xn > 0 && str[xn-1] >= '0' && str[xn-1] <= '9' && xdig < 5) {
        /* Produce digits */
        x += (str[xn-1] - '0')*xmul;
        xmul *= 10;
        ++xdig;
        --xn;
    }
    if (xdig > 4 || xdig < 3) {
        /* Coordinates cannot have more than 4 digits */
        return false;
    }
    if (xn > 0 && std::strchr(PUNCT, str[xn-1]) == 0) {
        /* Coordinates not preceded by proper punctuation */
        return false;
    }

    /* Locate Y after n */
    int y = 0;
    int ydig = 0;
    string_t::size_type yn = n+1;
    while (yn < str.size() && str[yn] == ' ') {
        /* Skip space */
        ++yn;
    }
    while (yn < str.size() && str[yn] >= '0' && str[yn] <= '9' && ydig < 5) {
        /* Produce digits */
        y *= 10;
        y += str[yn] - '0';
        ++ydig;
        ++yn;
    }
    if (ydig > 4 || ydig < 3) {
        /* Coordinates cannot have more than 4 digits */
        return false;
    }
    if (yn < str.size() && std::strchr(PUNCT, str[yn]) == 0) {
        /* Coordinates not followed by proper punctuation */
        return false;
    }

    /* Reject invalid values */
    if (x == 0 || y == 0) {
        return false;
    }

    /* Success */
    pt.x = x;
    pt.y = y;
    start = xn;
    end = yn;
    return true;
}


/** Find an X,Y pair within a string.
    \param str [in] String
    \param pt  [out] Parsed coordinate
    \return true if found, false if no valid coordinate pair in string */
static bool
findXY(const string_t& str, GPoint& pt)
{
    string_t::size_type start, end;
    for (string_t::size_type n = str.find(','); n != string_t::npos; n = str.find(',', n+1)) {
        if (findXY(str, n, start, end, pt))
            return true;
    }
    return false;
}


/** Check for message header. Header lines must fulfill the regexp
    "[A-Z]+ *:", i.e. any single word followed by optional spaces and a
    colon. We want to recognize
      "TO: race"     (THost)
      "TO  : race"   (PHost, English)
      "An  : race"   (PHost, German)
      "CC: race"     (PCC)
    We do not need to recognize "<<<Universal Message>>>", this is handled
    on the outside. We do not need to handle "<CC:" and the blank line
    between host's headers and ours; we see the message in
    "cooked" format after these idiosyncrasies have been resolved.

    \todo This does not recognize Estonian
      " SAAJA: race"
    but I consider Estonian in error here. This does not recognize
    Russian, which uses Cyrillic letters in the headers.

    Original: readmsg.pas, IsHeader */
static bool
isHeader(const string_t& line)
{
    using std::isalpha;
    string_t::size_type n = 0;
    while (n < line.size() && isalpha(uint8_t(line[n])))
        ++n;
    if (n == 0)
        return false;
    while (n < line.size() && line[n] == ' ')
        ++n;
    return (n < line.size() && line[n] == ':');
}

/************************** WMessageActionPanel **************************/

/** Constructor.
    \param mbox Mailbox to display */
WMessageActionPanel::WMessageActionPanel(GMessagebox& mbox)
    : mbox(mbox),
      conn_browse(mbox.sig_sel_changed.add(this, &WMessageActionPanel::onBrowse)),
      btn_prev(UTF_UP_ARROW, SDLK_PAGEUP),
      btn_next(UTF_DOWN_ARROW, SDLK_PAGEDOWN),
      btn_goto("G", 'g'),
      btn_reply("R", 'r'),
      btn_xy("X", 'x'),
      btn_redirect("T", 't'),
      btn_edit("E", 'e'),
      btn_delete("Del", SDLK_DELETE),
      btn_confirm("C", 'c'),
      btn_search("S", 's'),
      btn_write("W", 'w'),
      btn_forward("F", 'f'),
      have_goto(false),
      have_reply(false),
      have_xy(false),
      have_outgoing(false),
      have_confirm(false)
{ }

/** Destructor. */
WMessageActionPanel::~WMessageActionPanel()
{ }

/** Update after change. This evaluates the message and fills in the
    have_xxx fields and their parameters. It then displays the widgets
    as required.  */
void
WMessageActionPanel::onBrowse()
{
    // Clear current
    have_goto = have_reply = have_xy = have_outgoing = have_confirm = false;

    // Get message
    string_t msg = mbox.getText(mbox.getCurrent());

    // Type-specific controls
    updateMessageControls();

    // Check X,Y
    have_xy = findXY(msg, xy_pos);
    if (!have_xy && have_goto && goto_obj.getScreenId() == WScreenId::Starchart) {
        have_xy = true;
        xy_pos = goto_obj.getPos();
    }

    // Hide the 'Go to' button if it's the same as the 'XY' button
    hide_goto = (have_goto && have_xy && goto_obj.getScreenId() == WScreenId::Starchart && goto_obj.getPos() == xy_pos);

    updateWidgets();
    drawWidget();
}

/** Show a widget. Makes it visible if it is hidden, moves it if it is already visible.
    \param w Widget to show
    \param r Position */
void
WMessageActionPanel::showWidget(UIBaseWidget& w, GfxRect r)
{
    if (w.getParent() == 0) {
        add(w, r);
    } else {
        w.setExtent(r);
    }
}

/** Hide a widget.
    \param w Widget to hide */
void
WMessageActionPanel::hideWidget(UIBaseWidget& w)
{
    if (w.getParent() != 0) {
        removeChildWidget(w);
    }
}

/** Update widgets. This computes all child widgets' positions and
    shows or hides them as needed. */
void
WMessageActionPanel::updateWidgets()
{
    int x = getExtent().x + 5;
    int y = getExtent().y + 5;

    // Arrows
    showWidget(btn_prev, GfxRect(x, y, 20, 20));
    showWidget(btn_next, GfxRect(x + getExtent().w - 30, y, 20, 20));
    y += 25;

    // 'g' button
    if (have_goto && !hide_goto) {
        showWidget(btn_goto, GfxRect(x, y, 20, 20));
        y += 25;
    } else {
        hideWidget(btn_goto);
    }

    // 'r' button
    if (have_reply) {
        showWidget(btn_reply, GfxRect(x, y, 20, 20));
        y += 25;
    } else {
        hideWidget(btn_reply);
    }

    // 'x' button
    if (have_xy) {
        showWidget(btn_xy, GfxRect(x, y, 20, 20));
        y += 25;
    } else {
        hideWidget(btn_xy);
    }

    // 'c' button
    if (have_confirm) {
        showWidget(btn_confirm, GfxRect(x, y, 20, 20));
        y += 25;
    } else {
        hideWidget(btn_confirm);
    }

    y = getExtent().getBottomY();
    y -= 25;
    showWidget(btn_write, GfxRect(x, y, 20, 20));
    y -= 25;
    showWidget(btn_search, GfxRect(x, y, 20, 20));
    y -= 25;
    showWidget(btn_forward, GfxRect(x, y, 20, 20));

    // outgoing controls
    if (have_outgoing) {
        y -= 25;
        showWidget(btn_delete, GfxRect(x, y, 35, 20));
        y -= 25;
        showWidget(btn_redirect, GfxRect(x, y, 20, 20));
        y -= 25;
        showWidget(btn_edit, GfxRect(x, y, 20, 20));
    } else {
        hideWidget(btn_edit);
        hideWidget(btn_redirect);
        hideWidget(btn_delete);
    }
}

/** Display content. This draws the background panel and the labels
    for the buttons. */
void
WMessageActionPanel::drawContent(GfxCanvas& can)
{
    drawTiledArea(can, getExtent(), requestPixmap("bluetile"), COLOR_DARKBLUE, 0);
    drawFrameUp(can, getExtent());

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);

    int x = getExtent().x + 30;
    int y = getExtent().y + 5;
    int w = getExtent().w - 30;

    if (getMessageFilter().isFiltered(mbox.getHeading(mbox.getCurrent()))) {
        ctx.useFont(*fonts[FONT_NORMAL]);
        ctx.setColor(COLOR_GRAY);
    } else {
        ctx.useFont(*fonts[FONT_NORMAL_BOLD]);
        ctx.setColor(COLOR_WHITE);
    }
    ctx.setTextAlign(1, 1);
    outTextF(ctx, getExtent().x + getExtent().w/2, y+10, w-30, format("%d/%d", mbox.getCurrent()+1, mbox.getCount()));
    y += 25;

    ctx.setTextAlign(0, 0);
    if (have_goto && !hide_goto) {
        ctx.setColor(COLOR_GRAY);
        ctx.useFont(*fonts[FONT_SMALL]);
        if (goto_obj.getScreenId() == WScreenId::Starchart) {
            GPoint pt = goto_obj.getPos();
            outTextF(ctx, x, y, w, _("Go to map at"));
            ctx.setColor(COLOR_WHITE);
            ctx.useFont(*fonts[FONT_SMALL_BOLD]);
            outTextF(ctx, x, y+10, w, format("(%d,%d)", pt.x, pt.y));
        } else {
            outTextF(ctx, x, y, w, _("Go to"));
            ctx.setColor(COLOR_WHITE);
            ctx.useFont(*fonts[FONT_SMALL_BOLD]);
            outTextF(ctx, x, y+10, w, goto_obj.toString());
        }
        y += 25;
    }
    if (have_reply) {
        ctx.useFont(*fonts[FONT_SMALL]);
        ctx.setColor(COLOR_GRAY);
        outTextF(ctx, x, y, w, _("Reply to"));
        ctx.setColor(COLOR_WHITE);
        ctx.useFont(*fonts[FONT_SMALL_BOLD]);
        if (reply_race == 0) {
            outTextF(ctx, x, y+10, w, _("all players"));
        } else if (reply_race == 12) {
            outTextF(ctx, x, y+10, w, _("Host"));
        } else {
            outTextF(ctx, x, y+10, w, player_racenames.getShortName(reply_race));
        }
        y += 25;
    }
    if (have_xy) {
        ctx.setColor(COLOR_GRAY);
        ctx.useFont(*fonts[FONT_SMALL]);
        outTextF(ctx, x, y, w, _("Go to map at"));
        ctx.setColor(COLOR_WHITE);
        ctx.useFont(*fonts[FONT_SMALL_BOLD]);
        outTextF(ctx, x, y+10, w, format("(%d,%d)", xy_pos.x, xy_pos.y));
        y += 25;
    }
    if (have_confirm) {
        ctx.setColor(COLOR_GRAY);
        ctx.useFont(*fonts[FONT_SMALL]);
        outTextF(ctx, x, y+5, w, _("Confirm"));
        y += 25;
    }

    ctx.setColor(COLOR_GRAY);
    ctx.useFont(*fonts[FONT_SMALL]);

    y = getExtent().getBottomY();
    y -= 25;
    outTextF(ctx, x, y+5, w, _("Write to file"));
    y -= 25;
    outTextF(ctx, x, y+5, w, _("Search..."));
    y -= 25;
    outTextF(ctx, x, y+5, w, _("Forward..."));
    if (have_outgoing) {
        y -= 25;
        outTextF(ctx, x+15, y+5, w, _("Delete"));
        y -= 25;
        outTextF(ctx, x, y+5, w, _("To..."));
        y -= 25;
        outTextF(ctx, x, y+5, w, _("Edit..."));
        y += 25;
    }

    can.update(getExtent());
}

void
WMessageActionPanel::getLayoutInfo(LayoutInfo& info)
{
    // 1 em FONT_TITLE     = 18 px, so we need 2.7 em for the buttons.
    // 1 digit FONT_NORMAL = 8 px, so we need 2.7 em for 6 digits.
    info.min_size = info.pref_size = font_sizes[FONT_TITLE].scaledBy(7, 8);
    info.flags = li_Flexible;
}

void
WMessageActionPanel::onResize(GfxRect /*oldsize*/, bool /*by_layout*/)
{
    /* Do not call the derived method. It would do the same thing we do for
       the top widgets, but not move the bottom widgets correctly. */
    updateWidgets();
}

bool
WMessageActionPanel::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case 'x':
            doGoToXY();
            return true;

         case 'g':
            doGoToObject();
            return true;

         case 'r':
            doReply();
            return true;

         case SDLK_RETURN:
            /* Do what I mean: Reply is default unless it would go to
               the host and we have other options. This is because
               PHost sends wormhole messages as (-h) by default. */
            /* FIXME: what do we do with object transfers? Probably
               the user doesn't want to reply to these. */
            if (haveReply() && !((haveObject() || haveXY()) && getReply() == 12)) {
                doReply();
            } else if (haveObject()) {
                doGoToObject();
            } else {
                doGoToXY();
            }
            return true;
        }
    }
    return UIWidget::handleEvent(event, second_pass);
}

/** Perform 'go to X/Y' operation. */
void
WMessageActionPanel::doGoToXY()
{
    if (haveXY()) {
        postGoToChart(IntUserInterfaceBinding::get(), getXY());
    }
}

/** Perform 'go to' operation. */
void
WMessageActionPanel::doGoToObject()
{
    if (haveObject()) {
        WScreenId id = getObject();
        if (id.getScreenId() == WScreenId::Starchart) {
            postGoToChart(IntUserInterfaceBinding::get(), id.getPos());
        } else {
            postGoToScreen(IntUserInterfaceBinding::get(), id.getScreenId(), id.getId());
        }
    }
}

/** Perform 'reply' operation. */
void
WMessageActionPanel::doReply()
{
    if (haveReply()) {
        // Split message into lines
        GMessageLines lines;
        splitMessage(lines, mbox.getText(mbox.getCurrent()));

        // Skip headers. First line always is (-foo).
        GMessageLines::size_type first = 1;
        while (first < lines.size()
               && (lines[first].size() == 0
                   || lines[first] == GOutbox::UNIVERSAL_TEXT
                   || isHeader(lines[first])))
        {
            ++first;
        }

        // Quote remainder.
        string_t quotedMessage;
        while (first < lines.size()) {
            quotedMessage += '>';
            if (lines[first].size() > 0 && lines[first][0] != '>')
                quotedMessage += ' ';
            quotedMessage += lines[first];
            quotedMessage += '\n';
            ++first;
        }

        // Edit and send the message.
        doSendMessage(quotedMessage, reply_race ? GPlayerSet(reply_race) : GPlayerSet::allUpTo(NUM_PLAYERS));
    }
}
