/**
  *  \file game/actions/baseship.cc
  *  \brief Starbase Action: Build Starship
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "game/actions/baseship.h"
#include "game/actions/preconditions.h"
#include "game/reg.h"
#include "game/spec.h"
#include "game/team.h"
#include "util/nls.h"

static void
canonicalizeBuildOrder(TBuildOrder& o)
{
    if (!o.torp_count)
        o.torp_type = 0;
    if (!o.beam_count)
        o.beam_type = 0;
    o.unused = 0;
}

static inline int
findBestEngine(const GPlanet& p)
{
    int e = 1;
    while (e < NUM_ENGINES && getEngine(e+1).getTech() <= p.getBaseTechLevel(tl_Engine))
        ++e;
    return e;
}

static inline int
findBestBeam(const GPlanet& p)
{
    int b = 1;
    while (b < NUM_BEAMS && getBeam(b+1).getTech() <= p.getBaseTechLevel(tl_Beam))
        ++b;
    return b;
}

static inline int
findBestTorpedo(const GPlanet& p)
{
    int t = 1;
    while (t < NUM_TORPS && getTorp(t+1).getTech() <= p.getBaseTechLevel(tl_Torpedo))
        ++t;
    return t;
}

/** Ship Build Action. */
GStarbaseBuildShipAction::GStarbaseBuildShipAction(GPlanet& planet)
    : GStarbaseBuildAction(planet),
      order(),
      use_parts_from_storage(true),
      use_tech_upgrade(true),
      is_ok(true)
{
    mustHavePlayedBase(planet);
    order = planet.getBaseBuildOrder();
    use_parts_from_storage = prepareBuildOrder(order, planet);

    /* Update is_ok by calling getCost() which will call perform() */
    getCost();
}

GStarbaseBuildShipAction::~GStarbaseBuildShipAction()
{ }

/** Choose whether parts from storage will be used. If enabled, ship
    building will use parts if possible. If disabled, everything will
    be built anew, even when there is already a matching part in
    storage. */
void
GStarbaseBuildShipAction::setUsePartsFromStorage(bool b)
{
    if (b != use_parts_from_storage) {
        use_parts_from_storage = b;
        sig_changed.raise();
    }
}

/** Choose whether tech level upgrades are allowed. If not, we'll
    refuse to build any ship that needs an upgrade. */
void
GStarbaseBuildShipAction::setUseTechUpgrade(bool b)
{
    if (b != use_tech_upgrade) {
        use_tech_upgrade = b;
        sig_changed.raise();
    }
}

GError
GStarbaseBuildShipAction::getError()
{
    if (!is_ok)
        return GError(GError::ePerm, _("This build order cannot be performed"));
    return GError();
}

void
GStarbaseBuildShipAction::setBuildOrder(TBuildOrder o)
{
    using std::memcmp;
    canonicalizeBuildOrder(o);
    if (memcmp(&o, &order, sizeof(o)) != 0) {
        order = o;
        sig_changed.raise();
    }
}

/** Do a tech upgrade. This accounts resources, and sets is_ok to
    false if anything doesn't want us to upgrade.
    \param tl     which level we're talking about
    \param a      actor whose setTech we ought to call
    \param need   required level */
void
GStarbaseBuildShipAction::doTechUpgrade(TTechLevel tl, GStarbaseBuildExecuter& a, int need)
{
    /* do we need an upgrade? */
    if (getPlanet().getBaseTechLevel(tl) >= need)
        return;

    // FIXME handle required tech > 10
    /* are we allowed to upgrade? */
    if (need > 6 && !allowRegistered(getPlanet().getOwner()))
        is_ok = false;

    /* we need a tech upgrade but do not want to. In particular, do
       not include upgrades in the calculation. */
    if (!use_tech_upgrade) {
        is_ok = false;
        return;
    }

    a.setTech(tl, need);
}

/** Compute how many parts we have to build.
    \param need   we need this many parts...
    \param have   ...and we have this many already in storage */
int
GStarbaseBuildShipAction::getBuildAmount(int need, int have)
{
    if (!use_parts_from_storage)
        return need;
    else if (need <= have)
        return 0;
    else
        return need - have;
}

/** Check whether this action is a change to an existing build order.
    \retval true there already is a different ship build order
    \retval false the starbase has no build order, or this is the same */
bool
GStarbaseBuildShipAction::isChange() const
{
    // Fetch and canonicalize both orders
    TBuildOrder baseOrder = getPlanet().getBaseBuildOrder();
    TBuildOrder newOrder  = order;
    canonicalizeBuildOrder(baseOrder);
    canonicalizeBuildOrder(newOrder);

    // Compare
    return baseOrder.hull != 0
        && (baseOrder.hull != newOrder.hull
            || baseOrder.engine != newOrder.engine
            || baseOrder.beam_type != newOrder.beam_type
            || baseOrder.beam_count != newOrder.beam_count
            || baseOrder.torp_type != newOrder.torp_type
            || baseOrder.torp_count != newOrder.torp_count);
}

/** Gegenerate all orders, and update validity status. We have some
    more validity checks than just 'sufficient funds':
    - can we upgrade tech?
    - did the user choose no tech upgrade but we would need one? */
void
GStarbaseBuildShipAction::perform(GStarbaseBuildExecuter& a)
{
    int n;
    is_ok = true;

    // first, attempt to build one hull
    if (int slot = getTruehullSlot(getPlanet().getOwner(), order.hull)) {
        // we can build it
        if ((n = getBuildAmount(1, getPlanet().getBaseHullStoreSlot(slot))) != 0) {
            doTechUpgrade(tl_Hull, a, getHull(order.hull).getTech());
        }
        a.setHullSlot(slot, getPlanet().getBaseHullStoreSlot(slot) + n, 1 - n);
    } else {
        // we can not build it.
        is_ok = false;
        doTechUpgrade(tl_Hull, a, getHull(order.hull).getTech());
        a.accountHull(order.hull, 1, 0);
    }

    // build the engines
    if ((n = getBuildAmount(getHull(order.hull).getNumEngines(),
                            getPlanet().getBaseEngineStore(order.engine))) != 0)
    {
        doTechUpgrade(tl_Engine, a, getEngine(order.engine).getTech());
    }
    a.setEngine(order.engine, getPlanet().getBaseEngineStore(order.engine) + n, getHull(order.hull).getNumEngines() - n);

    // build the arms
    if (order.beam_type > 0 && order.beam_count > 0) {
        if ((n = getBuildAmount(order.beam_count, getPlanet().getBaseBeamStore(order.beam_type))) != 0) {
            doTechUpgrade(tl_Beam, a, getBeam(order.beam_type).getTech());
        }
        a.setBeam(order.beam_type, getPlanet().getBaseBeamStore(order.beam_type) + n, order.beam_count - n);
    }

    if (order.torp_count > 0 && order.torp_type > 0) {
        if ((n = getBuildAmount(order.torp_count, getPlanet().getBaseLauncherStore(order.torp_type))) != 0) {
            doTechUpgrade(tl_Torpedo, a, getTorp(order.torp_type).getTech());
        }
        a.setLauncher(order.torp_type, getPlanet().getBaseLauncherStore(order.torp_type) + n, order.torp_count - n);
    }
}

/** Prepare a build order.
    \param o [in/out] The build order
    \param p [in] Planet
    \retval true we're re-using the base's build order
    \retval false this is a new build order */
bool
GStarbaseBuildShipAction::prepareBuildOrder(TBuildOrder& o, const GPlanet& p)
{
    if (!o.hull) {
        // no pending build order. Invent one.
        o.hull = getTruehull(p.getOwner(), 1);
        o.beam_count = getHull(o.hull).getMaxBeams();
        o.torp_count = getHull(o.hull).getMaxLaunchers();
        o.engine = findBestEngine(p);
        o.beam_type = findBestBeam(p);
        o.torp_type = findBestTorpedo(p);
        return false;
    } else {
        // use existing build order
        if (!o.beam_count) {
            o.beam_type = findBestBeam(p);
        }
        if (!o.torp_count) {
            o.torp_type = findBestTorpedo(p);
        }
        return true;
    }
}
