/**
  *  \file game/alliance.h
  *  \brief Alliance Handling Core
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_ALLIANCE_H
#define PCC_V2_GAME_ALLIANCE_H

#include <vector>
#include "cpluslib/string.h"
#include "cpluslib/ptrvector.h"
#include "cpluslib/object.h"
#include "game/player.h"

class GAlliances;

/** Alliance level.
    Defines an alliance level.
    This object is immutable after creation. */
class GAllianceLevel {
 public:
    enum {
        IsOffer          = 1,        ///< This alliance level is the "offer an alliance" flag.
        NeedsOffer       = 2,        ///< This alliance level requires the "offer an alliance" flag set.
        IsEnemy          = 4,        ///< This alliance level is an "enemy" flag.
        AllowConditional = 8         ///< This alliance level allows conditional offers.
    };

    GAllianceLevel(const char* name, const char* id, int flags);
    ~GAllianceLevel();

    const string_t& getName() const;
    const string_t& getId() const;

    int getFlags() const;
    bool hasFlag(int fl) const;

 private:
    string_t name;
    string_t id;
    int flags;
};

/** Vector of alliance levels. */
typedef std::vector<GAllianceLevel> GAllianceLevels;

/** Alliance offer.
    This structure contains the status of an alliance offer. */
struct GAllianceOffer {
    enum OfferType {
        No,
        Unknown,
        Conditional,
        Yes
    };
    GPlayerArray<OfferType> theirOffer;
    GPlayerArray<OfferType> oldOffer;
    GPlayerArray<OfferType> newOffer;

    GAllianceOffer();

    void merge(const GAllianceOffer& other);

    static bool isOffer(OfferType type);
};

/** Vector of alliance offers. */
typedef std::vector<GAllianceOffer> GAllianceOffers;

/** Alliance handler.
    Provides a callback to update alliances. */
class GAllianceHandler : public Object {
 public:
    virtual void postprocess(GAlliances&) = 0;
    virtual void handleChanges(const GAlliances&) = 0;
};

/** Alliances, everything together. */
class GAlliances {
 public:
    typedef size_t index_t;
    static const index_t nil = index_t(-1);

    GAlliances();
    explicit GAlliances(const GAlliances&);
    ~GAlliances();

    void postprocess();
    void addLevel(const GAllianceLevel& level);
    void addNewHandler(GAllianceHandler* handler);

    void copyFrom(GAlliances& other);

    const GAllianceLevels& getLevels() const;
    const GAllianceOffers& getOffers() const;

    index_t find(const string_t& id) const;
    const GAllianceLevel* getLevel(index_t index) const;
    const GAllianceOffer* getOffer(index_t index) const;
    GAllianceOffer* getMutableOffer(index_t index);

    bool isAny(int player, int flag, bool fromUs) const;
    void setAll(int player, int flag, bool set);

    void set(index_t index, int player, GAllianceOffer::OfferType type);

 private:
    void callHandlers();

    GAllianceLevels levels;
    GAllianceOffers offers;
    ptr_vector<GAllianceHandler> handlers;
};

/******************************** Inlines ********************************/

/** Get name of this alliance level. */
inline const string_t&
GAllianceLevel::getName() const
{
    return name;
}

/** Get identifier for this alliance level. */
inline const string_t&
GAllianceLevel::getId() const
{
    return id;
}

/** Get flags. */
inline int
GAllianceLevel::getFlags() const
{
    return flags;
}

/** Check for flags.
    \param fl Flags to check for */
inline bool
GAllianceLevel::hasFlag(int fl) const
{
    return (flags & fl) != 0;
}

/** Add a new handler.
    \param handler Newly-allocated alliance handler */
inline void
GAlliances::addNewHandler(GAllianceHandler* handler)
{
    handlers.push_back(handler);
}

/** Get description of all levels. */
inline const GAllianceLevels&
GAlliances::getLevels() const
{
    return levels;
}

/** Get description of all offers. */
inline const GAllianceOffers&
GAlliances::getOffers() const
{
    return offers;
}

#endif
