/**
  *  \file game/rangeset.cc
  *  \brief Set of Circular Ranges
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "game/rangeset.h"
#include "game/objtypes.h"
#include "game/obj.h"

GRangeSet::GRangeSet()
    : points(),
      min(0,0),
      max(0,0)
{ }

GRangeSet::~GRangeSet()
{ }

/** Add a single range.
    \param pt Center
    \param r Radius */
void
GRangeSet::add(GPoint pt, int r)
{
    if (r > 0) {
        // Seed min, max
        if (points.empty()) {
            min = max = pt;
        }

        // Include point in map
        PointMap_t::iterator it = points.find(pt);
        if (it != points.end()) {
            if (r > it->second) {
                it->second = r;
            }
        } else {
            points.insert(std::make_pair(pt, r));
        }

        // Update min, max
        min.x = std::min(min.x, pt.x - r);
        min.y = std::min(min.y, pt.y - r);
        max.x = std::max(max.x, pt.x + r);
        max.y = std::max(max.y, pt.y + r);
    }
}

/** Add object type.
    \param type Type to iterate through
    \param playerLimit Only include these players
    \param markedOnly true to only include marked objects
    \param r Radius */
void
GRangeSet::addObjectType(GObjectType& type, GPlayerSet playerLimit, bool markedOnly, int r)
{
    for (int i = type.findNextIndex(0); i != 0; i = type.findNextIndex(i)) {
        if (GMapObject* mo = dynamic_cast<GMapObject*>(&type.getObjectByIndex(i))) {
            if (playerLimit.contains(mo->getOwner()) && (!markedOnly || mo->isMarked()))
            {
                add(mo->getPos(), r);
            }
        }
    }
}

/** Clear. */
void
GRangeSet::clear()
{
    points.clear();
    min = max = GPoint(0, 0);
}

/** Check for emptiness. */
bool
GRangeSet::isEmpty() const
{
    return points.empty();
}

/** Get minimum point of bounding box. */
GPoint
GRangeSet::getMinimum() const
{
    return min;
}

/** Get maximum point of bounding box. */
GPoint
GRangeSet::getMaximum() const
{
    return max;
}

/** Get begin iterator. */
GRangeSet::Iterator_t
GRangeSet::begin() const
{
    return points.begin();
}

/** Get end iterator. */
GRangeSet::Iterator_t
GRangeSet::end() const
{
    return points.end();
}
