/**
  *  \file game/save.cc
  *  \brief Saving Game Data -- General Utilities
  *
  *  This module contains the basic routines for writing out object
  *  lists.
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/save.h"
#include "game/global.h"
#include "game/ship.h"
#include "io/io.h"
#include "game/checksum.h"
#include "game/planet.h"
#include "util/except.h"
#include "util/nls.h"
#include "io/streams.h"

/** Save ships to stream. Writes a "count" word plus ship records.
    \param univ      universe to save
    \param str       stream to save in
    \param src       source flag mask to use
    \param checksum  checksum word to update
    \param flags     flags (save_XXX) */
void
saveShips(const GUniverse& univ, Stream& str, const GPlayerSet src, uint32_t& checksum, const int flags)
{
    /* pass 1: count */
    TInt16 v = { 0 };
    for (int i = 1; i <= NUM_SHIPS; ++i)
        if (univ.getShip(i).getShipSource().containsAnyOf(src))
            ++v.value;
    storeStructureT(str, v);
    if ((flags & save_ChecksumCount) != 0)
        checksum += getChecksum(v.value);

    /* pass 2: data */
    for (int i = 1; i <= NUM_SHIPS; ++i) {
        if (univ.getShip(i).getShipSource().containsAnyOf(src)) {
            char buffer[TShip::size];
            TShip tmp;
            univ.getShip(i).getShipData(tmp);
            storeStructure(buffer, tmp);
            str.writeT(buffer, sizeof buffer);

            uint32_t csum = getChecksum(buffer, sizeof buffer);
            setControlEntry(ShipSection, i, csum);
            checksum += csum;
        }
    }
}

/** Save planets to stream. Writes a "count" word plus planet records.
    \param univ      universe to save
    \param str       stream to save in
    \param src       source flag mask to use (generally 1<<pid)
    \param checksum  checksum word to update
    \param flags     flags (save_XXX) */
void
savePlanets(const GUniverse& univ, Stream& str, const GPlayerSet src, uint32_t& checksum, const int flags)
{
    /* pass 1: count */
    TInt16 v = { 0 };
    for (int i = 1; i <= NUM_PLANETS; ++i)
        if (univ.getPlanet(i).getPlanetSource().containsAnyOf(src))
            ++v.value;
    storeStructureT(str, v);
    if ((flags & save_ChecksumCount) != 0)
        checksum += getChecksum(v.value);

    /* pass 2: data */
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        if (univ.getPlanet(i).getPlanetSource().containsAnyOf(src)) {
            char buffer[TPlanet::size];
            TPlanet tmp;
            univ.getPlanet(i).getPlanetData(tmp);
            storeStructure(buffer, tmp);
            str.writeT(buffer, sizeof buffer);

            uint32_t csum = getChecksum(buffer, sizeof buffer);
            setControlEntry(PlanetSection, i, csum);
            checksum += csum;
        }
    }
}

/** Save bases to stream. Writes a "count" word plus base records.
    \param univ      universe to save
    \param str       stream to save in
    \param src       source flag mask to use (generally 1<<pid)
    \param checksum  checksum word to update
    \param flags     flags (save_XXX) */
void
saveBases(const GUniverse& univ, Stream& str, const GPlayerSet src, uint32_t& checksum, const int flags)
{
    /* pass 1: count */
    TInt16 v = { 0 };
    for (int i = 1; i <= NUM_PLANETS; ++i)
        if (univ.getPlanet(i).getBaseSource().containsAnyOf(src))
            ++v.value;
    storeStructureT(str, v);
    if ((flags & save_ChecksumCount) != 0)
        checksum += getChecksum(v.value);

    /* pass 2: data */
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        if (univ.getPlanet(i).getBaseSource().containsAnyOf(src)) {
            char buffer[TStarbase::size];
            TStarbase tmp;
            univ.getPlanet(i).getBaseData(tmp);
            storeStructure(buffer, tmp);
            str.writeT(buffer, sizeof buffer);

            uint32_t csum = getChecksum(buffer, sizeof buffer);
            setControlEntry(BaseSection, i, csum);
            checksum += csum;
        }
    }
}
