/**
  *  \file int/if/planetif.cc
  *  \brief Interpreter Game Interface: Planets
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/if/planetif.h"
#include "cpluslib/misc.h"
#include "game/global.h"
#include "game/planet.h"
#include "int/compound.h"
#include "int/data.h"
#include "int/if/baseif.h"
#include "int/if/ifutil.h"
#include "int/if/playerif.h"
#include "int/interror.h"
#include "int/intglobal.h"
#include "int/value.h"
#include "game/tables.h"
#include "game/ship.h"
#include "game/team.h"
#include "cpluslib/format.h"
#include "int/if/objif.h"
#include "game/actions/planetbase.h"
#include "game/actions/planetstructs.h"
#include "int/exec.h"
#include "game/actions/planetsupplies.h"
#include "game/undo.h"
#include "int/args.h"
#include "game/cspec.h"
#include "int/contextio.h"
#include "int/if/shipif.h"
#include "int/process.h"
#include "int/propenum.h"
#include "game/cargotr.h"
#include "game/planetform.h"
#include "game/planettransfer.h"
#include "game/shiptransfer.h"
#include "game/actions/preconditions.h"
#include "int/if/mapmsgif.h"

/* @q Planet(sid:Int):Obj (Function, Context)
   Access planet (and starbase) properties.
   Use as
   | ForEach Planet Do ...
   or
   | With Planet(n) Do ...

   @diff This function was available for use in %With under the name %Planets() since PCC 1.0.6.
   Do not use the name %Planets in new code, it is not supported by PCC2; use %Planet instead.

   @see int:index:group:planetproperty|Planet Properties, int:index:group:planetcommand|Planet Commands
   @since PCC 1.0.18, PCC2 1.99.8 */
IntValue*
IFPlanetGet(IntArgBlock& args)
{
    int32_t id;
    args.checkArgumentCount(1);
    if (!checkIntArg(id, args.getNext(), 1, NUM_PLANETS))
        return 0;

    if (!haveDisplayedTurn())
        return 0;

    return new IntPlanetContext(id);
}

IntContext*
IFPlanetMake()
{
    IntPlanetContext* pc = new IntPlanetContext(0);
    return pc->initFirst();
}

int32_t
IFPlanetDim(int32_t dim)
{
    return dim == 0 ? 1 : NUM_PLANETS+1;
}

/***************************** Planet Methods ****************************/

static void
doStructureBuild(GPlanet& pl, IntExecutionContext& exc, IntValue* countVal, IntValue* flagVal, TPlanetaryStructure str)
{
    int32_t count, flag;
    if (!checkIntArg(count, countVal, -10000, +10000))
        return;

    flag = 0;
    checkFlagArg(flag, 0, flagVal, "N");

    GMultiTransaction mt;
    GPlanetTransfer& transfer         = mt.addNewSubtransaction(new GPlanetTransfer(pl));
    GPlanetBuildStructuresAction& bsa = mt.addNewSubtransaction(new GPlanetBuildStructuresAction(pl, transfer));
    bsa.setUndoInformation(getDisplayedTurn().getPreviousUniverse().getPlanet(pl.getId()));

    int built = bsa.addStructuresLimitCash(str, count);
    if (flag) {
        // We permit partial build, place remainder in BUILD.REMAINDER
        IntIntIntValue iv(count - built);
        exc.setVariable("BUILD.REMAINDER", &iv);
        mt.doCommit();
    } else {
        // We do not permit partial build, so refuse it
        if (built != count) {
            mt.doCancel();
            throw GError(GError::ePerm, GError::ePerm);
        } else {
            mt.doCommit();
        }
    }
}

/* @q AutoBuild (Planet Command)
   Perform a standard auto-build operation.
   @since PCC 1.0.5, PCC2 1.99.9 */
static void
IFPlanetAutoBuild(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(0);

    GPlanet& pl = static_cast<GPlanet&>(obj);
    GMultiTransaction mt;
    GPlanetTransfer& transfer         = mt.addNewSubtransaction(new GPlanetTransfer(pl));
    GPlanetBuildStructuresAction& bsa = mt.addNewSubtransaction(new GPlanetBuildStructuresAction(pl, transfer));
    bsa.doStandardAutoBuild();
    mt.doCommit();
}

/* @q BuildBase Optional flag:Bool (Planet Command)
   Build a starbase.
   If the parameter is specified as True or not at all, builds the base.
   If the parameter is specified as False, cancels a previous build order.
   @since PCC 1.0.5, PCC2 1.99.9 */
static void
IFPlanetBuildBase(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    // Check arguments
    bool doit = true;
    args.checkArgumentCount(0, 1);
    if (args.getNumArgs() > 0)                // FIXME: this is inconsistent in that a missing arg means true, not ignore
        if (!checkBoolArg(doit, args.getNext()))
            return;

    // Do it
    GPlanet& pl = static_cast<GPlanet&>(obj);
    GMultiTransaction mt;
    GPlanetTransfer& transfer = mt.addNewSubtransaction(new GPlanetTransfer(pl));
    mt.addNewSubtransaction(new GPlanetBuildStarbaseAction(pl, transfer, doit));
    mt.doCommit();
}

/* @q BuildBaseDefense amount:Int, Optional flag:Str (Planet Command)
   Build starbase defense.
   Build the the specified number of structures, or scraps them if %amount is negative.
   Fails if you don't own the planet, don't have the required resources,
   or if the new amount of structures is not allowed by the rules.

   If the %flag is specified as <tt>"n"</tt>, the command will not fail due to lacking resources.
   Instead, it will build as many structures as possible, and set the variable {Build.Remainder}
   to the amount not built.
   @see BuildBaseDefenseWait
   @since PCC 1.0.5, PCC2 1.99.9 */
static void
IFPlanetBuildBaseDefense(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(1, 2);
    IntValue* amount = args.getNext();
    IntValue* flag = args.getNext();
    doStructureBuild(static_cast<GPlanet&>(obj),
                     exc,
                     amount,
                     flag,
                     str_BaseDefense);
}

/* @q BuildDefense amount:Int, Optional flag:Str (Planet Command)
   Build defense posts.
   Build the the specified number of structures, or scraps them if %amount is negative.
   Fails if you don't own the planet, don't have the required resources,
   or if the new amount of structures is not allowed by the rules.

   If the %flag is specified as <tt>"n"</tt>, the command will not fail due to lacking resources.
   Instead, it will build as many structures as possible, and set the variable {Build.Remainder}
   to the amount not built.
   @see BuildDefenseWait
   @since PCC 1.0.5, PCC2 1.99.9 */
static void
IFPlanetBuildDefense(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(1, 2);
    IntValue* amount = args.getNext();
    IntValue* flag = args.getNext();
    doStructureBuild(static_cast<GPlanet&>(obj),
                     exc,
                     amount,
                     flag,
                     str_Defense);
}

/* @q BuildFactories amount:Int, Optional flag:Str (Planet Command)
   Build factories.
   Build the the specified number of structures, or scraps them if %amount is negative.
   Fails if you don't own the planet, don't have the required resources,
   or if the new amount of structures is not allowed by the rules.

   If the %flag is specified as <tt>"n"</tt>, the command will not fail due to lacking resources.
   Instead, it will build as many structures as possible, and set the variable {Build.Remainder}
   to the amount not built.
   @see BuildFactoriesWait
   @since PCC 1.0.5, PCC2 1.99.9 */
static void
IFPlanetBuildFactories(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(1, 2);
    IntValue* amount = args.getNext();
    IntValue* flag = args.getNext();
    doStructureBuild(static_cast<GPlanet&>(obj),
                     exc,
                     amount,
                     flag,
                     str_Factories);
}

/* @q BuildMines amount:Int, Optional flag:Str (Planet Command)
   Build mineral mines.
   Build the the specified number of structures, or scraps them if %amount is negative.
   Fails if you don't own the planet, don't have the required resources,
   or if the new amount of structures is not allowed by the rules.

   If the %flag is specified as <tt>"n"</tt>, the command will not fail due to lacking resources.
   Instead, it will build as many structures as possible, and set the variable {Build.Remainder}
   to the amount not built.
   @see BuildMinesWait
   @since PCC 1.0.5, PCC2 1.99.9 */
static void
IFPlanetBuildMines(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(1, 2);
    IntValue* amount = args.getNext();
    IntValue* flag = args.getNext();
    doStructureBuild(static_cast<GPlanet&>(obj),
                     exc,
                     amount,
                     flag,
                     str_Mines);
}

/* @q SellSupplies amount:Int, Optional flags:Str (Planet Command)
   Sell or buy supplies.
   Sells the specified number of supplies (for one megacredit each), or buys supplies if %count is negative.
   You can only buy back supplies you sold this turn.
   Fails if you don't own the planet, or the rules forbid you to buy/sell the specified amount
   (because you don't have enough, maybe).

   Optionally, you can specify the flag "n", as in
   | SellSupplies 1000, "n"
   When you can't sell/buy the specified amount, this will sell as much as possible instead of failing.
   The variable {Build.Remainder} will be set to the amount that was not sold.
   For example, if the planet on which you run the above command only has 650 supplies,
   %Build.Remainder will be set to 350.
   @since PCC 1.0.19, PCC2 1.99.9 */
static void
IFPlanetSellSupplies(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& exc, IntArgBlock& args)
{
    // Fetch arguments
    args.checkArgumentCount(1, 2);

    int32_t amount;
    if (!checkIntArg(amount, args.getNext()))
        return;

    int32_t flag = 0;
    checkFlagArg(flag, 0, args.getNext(), "N");

    // Get planet
    GPlanet& pl = static_cast<GPlanet&>(obj);

    // Do it
    int32_t undid;
    if (amount == 0) {
        // No change
        undid = 0;
    } else if (amount > 0) {
        // Sell supplies. We have a transaction for that.
        GPlanetSellSuppliesAction action(pl);
        action.setAmount(amount);
        undid = amount - action.getAmount();
        if (!flag && undid != 0) {
            action.doCancel();
            throw GError(GError::ePerm, GError::ePerm);
        }
        action.doCommit();
    } else {
        // Buy supples. We do not yet have a transaction for that. FIXME.
        mustBePlayed(pl);

        // Compute how much we can buy
        GUndoInformation undo;
        undo.computeInformation(getDisplayedTurn(), pl.getId());
        amount = -amount;

        int32 supplies = undo.getSuppliesAllowedToBuy();
        int32 money    = pl.getCargo(el_Money);
        int32 maximum  = supplies > money ? money : supplies;

        if (amount > maximum) {
            undid = maximum - amount;     // should be negative
            if (flag)
                amount = maximum;
            else
                throw GError(GError::ePerm, GError::ePerm);
        } else {
            undid = 0;
        }

        // Do it.
        pl.changeCargoRaw(el_Supplies, +amount);
        pl.changeCargoRaw(el_Money,    -amount);
    }

    // Set variable
    if (flag) {
        IntIntIntValue iv(undid);
        exc.setVariable("BUILD.REMAINDER", &iv);
    }
}

/* @q SetColonistTax n:Int (Planet Command)
   Set colonist tax.
   @see Colonists.Tax
   @since PCC 1.0.5, PCC2 1.99.9 */
static void
IFPlanetSetColonistTax(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    setPlanetProperty(static_cast<GPlanet&>(obj), ippColonistTax, args.getNext());
}

/* @q SetComment s:Str (Planet Command)
   Set planet comment.
   @see Comment (Planet Property)
   @since PCC 1.0.5, PCC2 1.99.9 */
static void
IFPlanetSetComment(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    if (IntValue* v = args.getNext()) {
        IntStringValue sv(v->toString(false));
        int pid = obj.getId();
        getPlanetProperties(pid).set(sp_Comment, &sv);
        univ.ty_any_planets.getObjectByIndex(pid).markDirty();
    }
}

/* @q SetFCode fc:Str (Planet Command)
   Set planet friendly code.
   @see FCode (Planet Property)
   @since PCC 1.0.5, PCC2 1.99.9 */
static void
IFPlanetSetFCode(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    setPlanetProperty(static_cast<GPlanet&>(obj), ippFCode, args.getNext());
}

/* @q SetNativeTax n:Int (Planet Command)
   Set native tax.
   @see Natives.Tax
   @since PCC 1.0.5, PCC2 1.99.9 */
static void
IFPlanetSetNativeTax(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    setPlanetProperty(static_cast<GPlanet&>(obj), ippNativeTax, args.getNext());
}

/* Documented in shipif.cc, IFShipCargoTransfer */
static void
IFPlanetCargoTransfer(GObject& obj, GUniverse& univ, IntExecutionContext& exc, IntArgBlock& args)
{
    // Args are cargospec, ship Id, OSNJ flags + proxy ID
    // FIXME: proxy
    string_t cargospec;
    int32_t target_sid;
    int32_t flags = 0;
    args.checkArgumentCount(2, 3);
    if (!checkStringArg(cargospec, args.getNext()))
        return;
    if (!checkIntArg(target_sid, args.getNext(), 1, NUM_SHIPS))
        return;
    checkFlagArg(flags, 0, args.getNext(), "OSN");

    // Validate
    GPlanet& pl     = static_cast<GPlanet&>(obj);
    GShip& target   = univ.getShip(target_sid);
    if (!pl.isVisible() || !target.isVisible()) {
        throw IntError("Ship does not exist");
    }
    if (target.getPos() != pl.getPos()) {
        throw IntError(GError::ePos);
    }
    if (target.getShipKind() != GShip::CurrentShip) {
        throw IntError(GError::eNotPlaying);
    }

    // Build the transfer structures
    GCargoTransfer tx;
    tx.addNewContainer(makePlanetTransfer(univ, pl.getId(), target));
    tx.addNewContainer(new GShipTransfer(target));
    doScriptTransfer(exc, tx, cargospec, 0, flags);
    tx.doCommit();
}

/* @q AutoTaxColonists (Planet Command)
   Auto-tax for colonists.
   @since PCC2 1.99.15 */
static void
IFPlanetAutoTaxColonists(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(0);

    GPlanet& pl = static_cast<GPlanet&>(obj);
    if (!pl.isPlayable(GObject::Playable)) {
        throw IntError(GError::eNotPlaying);
    }

    mp16_t val = getColonistSafeTax(pl, pl.getStructures(str_Mines) + pl.getStructures(str_Factories));
    if (val.isKnown()) {
        pl.setColonistTax(val);
    }
}

/* @q AutoTaxNatives (Planet Command)
   Auto-tax for natives.
   @since PCC2 1.99.15 */
static void
IFPlanetAutoTaxNatives(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(0);

    GPlanet& pl = static_cast<GPlanet&>(obj);
    if (!pl.isPlayable(GObject::Playable)) {
        throw IntError(GError::eNotPlaying);
    }
    mp16_t val = getNativeSafeTax(pl, pl.getStructures(str_Mines) + pl.getStructures(str_Factories));
    if (val.isKnown()) {
        pl.setNativeTax(val);
    }
}

/*********************** IntPlanetArrayProperty **********************/

namespace {
    class IntPlanetArrayProperty : public IntIndexableValue {
     public:
        IntPlanetArrayProperty(const GPlanet& planet, IntPlanetProperty property);

        // IntIndexableValue:
        IntValue* get(IntArgBlock& args);
        void set(IntArgBlock& args, IntValue* value);
        int32_t getDimension(int32_t dim);
        IntContext* makeFirstContext();

        // IntValue:
        IntPlanetArrayProperty* clone();
        string_t toString(bool readable);
        void store(IntTagNode& sv, Stream& aux);

     private:
        const GPlanet& planet;
        const IntPlanetProperty property;
    };
}

IntPlanetArrayProperty::IntPlanetArrayProperty(const GPlanet& planet, const IntPlanetProperty property)
    : IntIndexableValue(),
      planet(planet),
      property(property)
{ }

IntValue*
IntPlanetArrayProperty::get(IntArgBlock& args)
{
    switch (property) {
     case ippScore:
        /* Documented in shipif.cc */
        {
            args.checkArgumentCount(1);
            int32_t id;
            if (!checkIntArg(id, args.getNext(), 0, 0x7FFF)) {
                return 0;
            }
            int32_t score = planet.unit_scores.getScore(planet_score_definitions.lookupScore(id));
            if (score < 0) {
                return 0;
            } else {
                return makeIntValue(score);
            }
        }

     default:
        return 0;
    }
}

void
IntPlanetArrayProperty::set(IntArgBlock& /*args*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

int32_t
IntPlanetArrayProperty::getDimension(int32_t /*dim*/)
{
    return 0;
}

IntContext*
IntPlanetArrayProperty::makeFirstContext()
{
    throw IntError::typeError(IntError::tyIterable);
}

IntPlanetArrayProperty*
IntPlanetArrayProperty::clone()
{
    return new IntPlanetArrayProperty(planet, property);
}

string_t
IntPlanetArrayProperty::toString(bool /*readable*/)
{
    return "#<array>";
}

void
IntPlanetArrayProperty::store(IntTagNode& /*sv*/, Stream& /*aux*/)
{
    throw IntError::notSerializable();
}

/**************************** IntPlanetContext ***************************/

namespace { enum { PlanetProperty, BaseProperty, OwnerProperty, PlanetMethod }; }

static const TIntMapping planet_mapping[] = {
    { "AUTOBUILD",                 ipmAutoBuild,         PlanetMethod,   thProcedure },
    { "AUTOTAXCOLONISTS",          ipmAutoTaxColonists,  PlanetMethod,   thProcedure },
    { "AUTOTAXNATIVES",            ipmAutoTaxNatives,    PlanetMethod,   thProcedure },
    { "BASE",                      ippBaseStr,           PlanetProperty, thString },
    { "BASE.BUILDING",             ippBaseBuildFlag,     PlanetProperty, thBool },
    { "BASE.YESNO",                ippBaseFlag,          PlanetProperty, thBool },
    { "BUILD",                     ibpBuildHullName,     BaseProperty,   thString },
    { "BUILD.BEAM$",               ibpBuildBeam,         BaseProperty,   thInt },
    { "BUILD.BEAM.COUNT",          ibpBuildBeamCount,    BaseProperty,   thInt },
    { "BUILD.ENGINE$",             ibpBuildEngine,       BaseProperty,   thInt },
    { "BUILD.HULL$",               ibpBuildHull,         BaseProperty,   thInt },
    { "BUILD.QPOS",                ibpBuildQueuePos,     BaseProperty,   thInt },
    { "BUILD.SHORT",               ibpBuildHullShort,    BaseProperty,   thInt },
    { "BUILD.TORP$",               ibpBuildTorp,         BaseProperty,   thInt },
    { "BUILD.TORP.COUNT",          ibpBuildTorpCount,    BaseProperty,   thInt },
    { "BUILD.YESNO",               ibpBuildFlag,         BaseProperty,   thBool },
    { "BUILDBASE",                 ipmBuildBase,         PlanetMethod,   thProcedure },
    { "BUILDBASEDEFENSE",          ipmBuildBaseDefense,  PlanetMethod,   thProcedure },
    { "BUILDBEAMS",                ipmBuildBeams,        PlanetMethod,   thProcedure },
    { "BUILDDEFENSE",              ipmBuildDefense,      PlanetMethod,   thProcedure },
    { "BUILDENGINES",              ipmBuildEngines,      PlanetMethod,   thProcedure },
    { "BUILDFACTORIES",            ipmBuildFactories,    PlanetMethod,   thProcedure },
    { "BUILDFIGHTERS",             ipmBuildFighters,     PlanetMethod,   thProcedure },
    { "BUILDHULLS",                ipmBuildHulls,        PlanetMethod,   thProcedure },
    { "BUILDLAUNCHERS",            ipmBuildLaunchers,    PlanetMethod,   thProcedure },
    { "BUILDMINES",                ipmBuildMines,        PlanetMethod,   thProcedure },
    { "BUILDSHIP",                 ipmBuildShip,         PlanetMethod,   thProcedure },
    { "BUILDTORPS",                ipmBuildTorps,        PlanetMethod,   thProcedure },
    { "CARGOTRANSFER",             ipmCargoTransfer,     PlanetMethod,   thProcedure },
    { "COLONISTS",                 ippColonists,         PlanetProperty, thInt },
    { "COLONISTS.CHANGE",          ippColonistChangeStr, PlanetProperty, thString },
    { "COLONISTS.CHANGE$",         ippColonistChange,    PlanetProperty, thInt },
    { "COLONISTS.HAPPY",           ippColonistHappyStr,  PlanetProperty, thString },
    { "COLONISTS.HAPPY$",          ippColonistHappy,     PlanetProperty, thInt },
    { "COLONISTS.SUPPORTED",       ippColonistSupported, PlanetProperty, thInt },
    { "COLONISTS.TAX",             ippColonistTax,       PlanetProperty, thInt },
    { "COLONISTS.TAX.INCOME",      ippColonistTaxIncome, PlanetProperty, thInt },
    { "DAMAGE",                    ibpBaseDamage,        BaseProperty,   thInt },
    { "DEFENSE",                   ippDefense,           PlanetProperty, thInt },
    { "DEFENSE.BASE",              ibpBaseDefense,       BaseProperty,   thInt },
    { "DEFENSE.BASE.MAX",          ibpBaseDefenseMax,    BaseProperty,   thInt },
    { "DEFENSE.BASE.WANT",         ippBaseDefenseWanted, PlanetProperty, thInt },
    { "DEFENSE.MAX",               ippDefenseMax,        PlanetProperty, thInt },
    { "DEFENSE.WANT",              ippDefenseWanted,     PlanetProperty, thInt },
    { "DENSITY.D",                 ippDensityD,          PlanetProperty, thInt },
    { "DENSITY.M",                 ippDensityM,          PlanetProperty, thInt },
    { "DENSITY.N",                 ippDensityN,          PlanetProperty, thInt },
    { "DENSITY.T",                 ippDensityT,          PlanetProperty, thInt },
    { "FACTORIES",                 ippFactories,         PlanetProperty, thInt },
    { "FACTORIES.MAX",             ippFactoriesMax,      PlanetProperty, thInt },
    { "FACTORIES.WANT",            ippFactoriesWanted,   PlanetProperty, thInt },
    { "FCODE",                     ippFCode,             PlanetProperty, thString },
    { "FIGHTERS",                  ibpBaseFighters,      BaseProperty,   thInt },
    { "FIGHTERS.MAX",              ibpBaseFightersMax,   BaseProperty,   thInt },
    { "FIXSHIP",                   ipmFixShip,           PlanetMethod,   thProcedure },
    { "GROUND.D",                  ippGroundD,           PlanetProperty, thInt },
    { "GROUND.M",                  ippGroundM,           PlanetProperty, thInt },
    { "GROUND.N",                  ippGroundN,           PlanetProperty, thInt },
    { "GROUND.T",                  ippGroundT,           PlanetProperty, thInt },
    { "ID",                        ippId,                PlanetProperty, thInt },
    { "INDUSTRY",                  ippIndustry,          PlanetProperty, thString },
    { "INDUSTRY$",                 ippIndustryCode,      PlanetProperty, thInt },
    { "LEVEL",                     ippLevel,             PlanetProperty, thInt },
    { "LOC.X",                     ippLocX,              PlanetProperty, thInt },
    { "LOC.Y",                     ippLocY,              PlanetProperty, thInt },
    { "MARK",                      ipmMark,              PlanetMethod,   thProcedure },
    { "MARKED",                    ippMarked,            PlanetProperty, thBool },
    { "MESSAGES",                  ippMessages,          PlanetProperty, thArray },
    { "MINED.D",                   ippMinedD,            PlanetProperty, thInt },
    { "MINED.M",                   ippMinedM,            PlanetProperty, thInt },
    { "MINED.N",                   ippMinedN,            PlanetProperty, thInt },
    { "MINED.STR",                 ippMinedStr,          PlanetProperty, thString},
    { "MINED.T",                   ippMinedT,            PlanetProperty, thInt },
    { "MINES",                     ippMines,             PlanetProperty, thInt },
    { "MINES.MAX",                 ippMinesMax,          PlanetProperty, thInt },
    { "MINES.WANT",                ippMinesWanted,       PlanetProperty, thInt },
    { "MISSION",                   ibpMissionName,       BaseProperty,   thString },
    { "MISSION$",                  ibpMission,           BaseProperty,   thInt },
    { "MONEY",                     ippMoney,             PlanetProperty, thInt },
    { "NAME",                      ippName,              PlanetProperty, thString },
    { "NATIVES",                   ippNatives,           PlanetProperty, thString },
    { "NATIVES.CHANGE",            ippNativeChangeStr,   PlanetProperty, thString },
    { "NATIVES.CHANGE$",           ippNativeChange,      PlanetProperty, thInt },
    { "NATIVES.GOV",               ippNativeGov,         PlanetProperty, thString },
    { "NATIVES.GOV$",              ippNativeGovCode,     PlanetProperty, thInt },
    { "NATIVES.HAPPY",             ippNativeHappyStr,    PlanetProperty, thString },
    { "NATIVES.HAPPY$",            ippNativeHappy,       PlanetProperty, thInt },
    { "NATIVES.RACE",              ippNativeRace,        PlanetProperty, thString },
    { "NATIVES.RACE$",             ippNativeRaceCode,    PlanetProperty, thInt },
    { "NATIVES.TAX",               ippNativeTax,         PlanetProperty, thInt },
    { "NATIVES.TAX.BASE",          ippNativeTaxBase,     PlanetProperty, thInt },
    { "NATIVES.TAX.INCOME",        ippNativeTaxIncome,   PlanetProperty, thInt },
    { "NATIVES.TAX.MAX",           ippNativeTaxMax,      PlanetProperty, thInt },
    { "ORBIT",                     ippOrbitingShips,     PlanetProperty, thInt },
    { "ORBIT.ENEMY",               ippOrbitingEnemies,   PlanetProperty, thInt },
    { "ORBIT.OWN",                 ippOrbitingOwn,       PlanetProperty, thInt },
    { "OWNER",                     iplShortName,         OwnerProperty,  thString },
    { "OWNER$",                    iplId,                OwnerProperty,  thInt },
    { "OWNER.ADJ",                 iplAdjName,           OwnerProperty,  thString },
    { "PLAYED",                    ippPlayed,            PlanetProperty, thBool },
    { "RECYCLESHIP",               ipmRecycleShip,       PlanetMethod,   thProcedure },
    { "SCORE",                     ippScore,             PlanetProperty, thArray },
    { "SELLSUPPLIES",              ipmSellSupplies,      PlanetMethod,   thProcedure },
    { "SETCOLONISTTAX",            ipmSetColonistTax,    PlanetMethod,   thProcedure },
    { "SETCOMMENT",                ipmSetComment,        PlanetMethod,   thProcedure },
    { "SETFCODE",                  ipmSetFCode,          PlanetMethod,   thProcedure },
    { "SETMISSION",                ipmSetMission,        PlanetMethod,   thProcedure },
    { "SETNATIVETAX",              ipmSetNativeTax,      PlanetMethod,   thProcedure },
    { "SETTECH",                   ipmSetTech,           PlanetMethod,   thProcedure },
    { "SHIPYARD",                  ibpShipyardStr,       BaseProperty,   thString },
    { "SHIPYARD.ACTION",           ibpShipyardAction,    BaseProperty,   thString },
    { "SHIPYARD.ID",               ibpShipyardId,        BaseProperty,   thInt },
    { "SHIPYARD.NAME",             ibpShipyardName,      BaseProperty,   thString },
    { "STORAGE.AMMO",              ibpAmmoStorage,       BaseProperty,   thArray },
    { "STORAGE.BEAMS",             ibpBeamStorage,       BaseProperty,   thArray },
    { "STORAGE.ENGINES",           ibpEngineStorage,     BaseProperty,   thArray },
    { "STORAGE.HULLS",             ibpHullStorage,       BaseProperty,   thArray },
    { "STORAGE.LAUNCHERS",         ibpLauncherStorage,   BaseProperty,   thArray },
    { "SUPPLIES",                  ippSupplies,          PlanetProperty, thInt },
    { "TASK",                      ippTask,              PlanetProperty, thBool },
    { "TASK.BASE",                 ippTaskBase,          PlanetProperty, thBool },
    { "TECH.BEAM",                 ibpBeamTech,          BaseProperty,   thInt },
    { "TECH.ENGINE",               ibpEngineTech,        BaseProperty,   thInt },
    { "TECH.HULL",                 ibpHullTech,          BaseProperty,   thInt },
    { "TECH.TORPEDO",              ibpTorpedoTech,       BaseProperty,   thInt },
    { "TEMP",                      ippTempStr,           PlanetProperty, thString },
    { "TEMP$",                     ippTemp,              PlanetProperty, thInt },
    { "TYPE",                      ippTypeStr,           PlanetProperty, thString },
    { "TYPE.SHORT",                ippTypeChar,          PlanetProperty, thString },
    { "UNMARK",                    ipmUnmark,            PlanetMethod,   thProcedure },
};

static const IntObjectProcedureValue::call_t planet_methods[] = {
    IFObjMark,                  // 0
    IFObjUnmark,                // 1
    IFPlanetSetComment,         // 2
    IFBaseFixShip,              // 3
    IFBaseRecycleShip,          // 4
    IFPlanetBuildBase,          // 5
    IFPlanetAutoBuild,          // 6
    IFPlanetBuildDefense,       // 7
    IFPlanetBuildFactories,     // 8
    IFPlanetBuildMines,         // 9
    IFPlanetSetColonistTax,     // 10
    IFPlanetSetNativeTax,       // 11
    IFPlanetSetFCode,           // 12
    IFBaseSetMission,           // 13
    IFPlanetBuildBaseDefense,   // 14
    IFBaseSetTech,              // 15
    IFBaseBuildFighters,        // 16
    IFBaseBuildEngines,         // 17
    IFBaseBuildHulls,           // 18
    IFBaseBuildLaunchers,       // 19
    IFBaseBuildBeams,           // 20
    IFBaseBuildTorps,           // 21
    IFPlanetSellSupplies,       // 22
    IFBaseBuildShip,            // 23
    IFPlanetCargoTransfer,      // 24
    IFPlanetAutoTaxColonists,   // 25
    IFPlanetAutoTaxNatives,     // 26
};

IntPlanetContext::IntPlanetContext(int pid)
    : IntObjectContext(current_anyplanet_type, pid)
{ }

IntPlanetContext::~IntPlanetContext()
{ }

static int32_t
lookupPlanetProperty(const IntNameQuery& name)
{
    static const int32_t nproperties = countof(planet_mapping);

    /* Check user-defined properties */
    IntVariableNames::index_t ix = planet_property_names.lookup(name);
    if (ix != planet_property_names.nil)
        return ix + nproperties;

    /* Check predefined properties */
    return lookupName(name, planet_mapping, nproperties);
}

int32_t
IntPlanetContext::lookup(const IntNameQuery& name)
{
    if (name.startsWith("PLANET.", 7)) {
        return lookupPlanetProperty(IntNameQuery(name, 7));
    } else {
        return lookupPlanetProperty(name);
    }
}

void
IntPlanetContext::set(int32_t index, IntValue* value)
{
    static const int32_t nproperties = countof(planet_mapping);
    if (index < nproperties) {
        /* Builtin property */
        GPlanet* pl = dynamic_cast<GPlanet*>(getObject());
        if (!pl) {
            throw IntError::notAssignable();
        }
        switch (planet_mapping[index].domain) {
         case PlanetProperty:
            setPlanetProperty(*pl, IntPlanetProperty(planet_mapping[index].index), value);
            break;
         case BaseProperty:
            setBaseProperty(*pl, IntBaseProperty(planet_mapping[index].index), value);
            break;
         default:
            throw IntError::notAssignable();
        }
    } else {
        /* User property */
        getPlanetProperties(getCurrentIndex()).set(index - nproperties, value);
        if (GObject* obj = getObject()) {
            obj->markDirty();
        }
    }
}

IntValue*
IntPlanetContext::get(int32_t index)
{
    static const int32_t nproperties = countof(planet_mapping);
    if (index < nproperties) {
        /* Builtin property */
        GPlanet* pl = dynamic_cast<GPlanet*>(getObject());
        if (!pl) {
            return 0;
        }
        switch (planet_mapping[index].domain) {
         case PlanetProperty:
            return getPlanetProperty(*pl, *getCurrentUniverse(), IntPlanetProperty(planet_mapping[index].index));
         case BaseProperty:
            return getBaseProperty(*pl, *getCurrentUniverse(), IntBaseProperty(planet_mapping[index].index));
         case OwnerProperty:
            if (pl->isOwnerKnown())
                return getPlayerProperty(pl->getOwner(), IntPlayerProperty(planet_mapping[index].index));
            else
                return 0;
         case PlanetMethod:
            return new IntObjectProcedureValue(*pl, *getCurrentUniverse(), planet_methods[planet_mapping[index].index]);
        }
        return 0;
    } else {
        /* User property */
        return IntValue::cloneOf(getPlanetProperties(getCurrentIndex())[index - nproperties]);
    }
}

IntContext*
IntPlanetContext::clone()
{
    return new IntPlanetContext(getCurrentIndex());
}

string_t
IntPlanetContext::toString(bool /*readable*/)
{
    return format("Planet(%d)", getCurrentIndex());
}

void
IntPlanetContext::store(IntTagNode& sv, Stream& /*aux*/)
{
    sv.tag   = Tag_Planet;
    sv.value = getCurrentIndex();
}

void
IntPlanetContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumNames(planet_property_names);
    acceptor.enumMapping(planet_mapping, countof(planet_mapping));
}

static int
countShips(GUniverse& univ, GPoint pt, GPlayerSet players)
{
    int n = 0;
    for (int sid = univ.ty_any_ships.findNextIndex(0); sid != 0; sid = univ.ty_any_ships.findNextIndex(sid)) {
        const GShip& sh = univ.ty_any_ships.getObjectByIndex(sid);
        if (pt == sh.getPos() && players.contains(sh.getOwner()))
            ++n;
    }
    return n;
}

IntValue*
getPlanetProperty(const GPlanet& pl, GUniverse& univ, IntPlanetProperty ipp)
{
    TAutobuildSettings abs;
    switch (ipp) {
     case ippBaseDefenseWanted:
        /* @q Defense.Base.Want:Int (Planet Property)
           Auto-build goal for starbase defense.
           @assignable */
        pl.getAutobuild(abs);
        return makeIntValue(abs.goal[str_BaseDefense]);
     case ippBaseFlag:
        /* @q Base.YesNo:Bool (Planet Property)
           True if this planet has a base. */
        return makeBoolValue(pl.hasBase());
     case ippBaseBuildFlag:
        /* @q Base.Building:Bool (Planet Property)
           True if this planet is building a base. */
        if (pl.hasFullPlanetData()) {
            return makeBoolValue(pl.isBuildingBase());
        } else {
            return 0;
        }
     case ippBaseStr:
        /* @q Base:Str (Planet Property)
           Starbase status, human-readable.
           One of
           - "present"
           - "being built"
           - "-" */
        if (pl.hasBase())
            return makeStringValue("present");
        else if (pl.isBuildingBase())
            return makeStringValue("being built");
        else
            return makeStringValue("-");
     case ippColonistChange:
        /* @q Colonists.Change$:Int (Planet Property)
           Colonist happiness change, numeric value. */
        return makeIntFromMaybe(getColonistChange(pl));
     case ippColonistChangeStr:
        /* @q Colonists.Change:Str (Planet Property)
           Colonist happiness change, text. */
        {
            mn32_t change = getColonistChange(pl);
            if (change.isKnown())
                return makeStringValue(getHappyChangeName(change));
            else
                return 0;
        }
     case ippColonistHappy:
        /* @q Colonists.Happy$:Int (Planet Property)
           Colonist happiness, numeric value. */
        return makeIntFromMaybe(pl.getColonistHappiness());
     case ippColonistHappyStr:
        /* @q Colonists.Happy:Str (Planet Property)
           Colonist happiness, text. */
        if (pl.getColonistHappiness().isKnown())
            return makeStringValue(getHappinessName(pl.getColonistHappiness()));
        else
            return 0;
     case ippColonistSupported:
        /* @q Colonists.Supported:Int (Planet Property)
           Maximum colonist clans supported by planet's climate.
           @since PCC 1.1.16, PCC2 1.99.8 */
        return makeIntFromMaybe(getMaxSupportedColonists(pl));
     case ippColonistTax:
        /* @q Colonists.Tax:Int (Planet Property)
           Colonist tax.
           @assignable
           @see SetColonistTax (Planet Command) */
        return makeIntFromMaybe(pl.getColonistTax());
     case ippColonistTaxIncome:
        /* @q Colonists.Tax.Income:Int (Planet Property)
           Tax income from colonists, megacredits.
           @since PCC2 1.99.15 */
        if (pl.isOwnerKnown() && pl.getOwner() > 0 && pl.getColonists().isKnown() && pl.getColonistTax().isKnown()) {
            return makeIntValue(getColonistDue(pl, pl.getColonistTax()));
        } else {
            return 0;
        }
     case ippColonists:
        /* @q Colonists:Int (Planet Property)
           Colonist population, number of clans. */
        return makeIntFromMaybe(pl.getColonists());
     case ippDefense:
        /* @q Defense:Int (Planet Property)
           Number of planetary defense posts. */
        return makeIntFromMaybe(pl.getStructures(str_Defense));
     case ippDefenseMax:
        /* @q Defense.Max:Int (Planet Property)
           Maximum number of planetary defense posts. */
        return makeIntFromMaybe(getMaxStructures(pl, str_Defense));
     case ippDefenseWanted:
        /* @q Defense.Want:Int (Planet Property)
           Auto-build goal for defense posts.
           @assignable */
        pl.getAutobuild(abs);
        return makeIntValue(abs.goal[str_Defense]);
     case ippDensityD:
        /* @q Density.D:Int (Planet Property)
           Density of Duranium in planet core. */
        return makeIntFromMaybe(pl.getOreDensity(el_Duranium));
     case ippDensityM:
        /* @q Density.M:Int (Planet Property)
           Density of Molybdenum in planet core. */
        return makeIntFromMaybe(pl.getOreDensity(el_Molybdenum));
     case ippDensityN:
        /* @q Density.N:Int (Planet Property)
           Density of Neutronium in planet core. */
        return makeIntFromMaybe(pl.getOreDensity(el_Neutronium));
     case ippDensityT:
        /* @q Density.T:Int (Planet Property)
           Density of Tritanium in planet core. */
        return makeIntFromMaybe(pl.getOreDensity(el_Tritanium));
     case ippFCode:
        /* @q FCode:Str (Planet Property)
           Friendly code.
           @assignable
           @see SetFCode (Planet Command) */
        if (pl.isFCodeKnown())
            return makeStringValue(pl.getFCode());
        else
            return 0;
     case ippFactories:
        /* @q Factories:Int (Planet Property)
           Number of factories on planet. */
        return makeIntFromMaybe(pl.getStructures(str_Factories));
     case ippFactoriesMax:
        /* @q Factories.Max:Int (Planet Property)
           Maximum number of factories on planet. */
        return makeIntFromMaybe(getMaxStructures(pl, str_Factories));
     case ippFactoriesWanted:
        /* @q Factories.Want:Int (Planet Property)
           Auto-build goal for factories.
           @assignable */
        pl.getAutobuild(abs);
        return makeIntValue(abs.goal[str_Factories]);
     case ippGroundD:
        /* @q Ground.D:Int (Planet Property)
           Amount of Duranium in ground, kilotons. */
        return makeIntFromMaybe(pl.getOreGround(el_Duranium));
     case ippGroundM:
        /* @q Ground.M:Int (Planet Property)
           Amount of Molybdenum in ground, kilotons. */
        return makeIntFromMaybe(pl.getOreGround(el_Molybdenum));
     case ippGroundN:
        /* @q Ground.N:Int (Planet Property)
           Amount of Neutronium in ground, kilotons. */
        return makeIntFromMaybe(pl.getOreGround(el_Neutronium));
     case ippGroundT:
        /* @q Ground.T:Int (Planet Property)
           Amount of Tritanium in ground, kilotons. */
        return makeIntFromMaybe(pl.getOreGround(el_Tritanium));
     case ippId:
        /* @q Id:Int (Planet Property)
           Planet Id. */
        return makeIntValue(pl.getId());
     case ippIndustry:
        /* @q Industry:Str (Planet Property)
           Planetary industry level, human-readable.
           @see Industry$ (Planet Property) */
        if (pl.getIndustryLevel().isKnown())
            return makeStringValue(getIndustryLevelName(pl.getIndustryLevel()));
        else
            return 0;
     case ippIndustryCode:
        /* @q Industry$:Int (Planet Property)
           Planetary industry level code.
           <table>
            <tr><th width="3" align="left">Ind$</th><th width="7" align="left">Ind</th></tr>
            <tr><td>0</td><td>Minimal</td></tr>
            <tr><td>1</td><td>Light</td></tr>
            <tr><td>2</td><td>Moderate</td></tr>
            <tr><td>3</td><td>Substantial</td></tr>
            <tr><td>4</td><td>Heavy</td></tr>
           </table> */
        return makeIntFromMaybe(pl.getIndustryLevel());
     case ippLevel:
        /* @q Level:Int (Planet Property)
           Planet's experience level.
           If the experience system is not enabled, or the level is not known, yields EMPTY. */
        {
            int16_t level = pl.unit_scores.getScore(planet_score_definitions.lookupScore(ScoreId_ExpLevel));
            if (level >= 0)
                return makeIntValue(level);
            else
                return 0;
        }
     case ippLocX:
        /* @q Loc.X:Int (Planet Property)
           Planet X location. */
        return makeIntValue(pl.getPos().x);
     case ippLocY:
        /* @q Loc.Y:Int (Planet Property)
           Planet Y location. */
        return makeIntValue(pl.getPos().y);
     case ippMarked:
        /* @q Marked:Bool (Planet Property)
           True if planet is marked. */
        return makeBoolValue(pl.isMarked());
     case ippMessages:
        /* @q Messages:Obj() (Planet Property)
           If this planet has any messages, this property is non-null and contains an array of messages.
           Individual messages have the same form as the inbox messages (InMsg()).
           @see int:index:group:incomingmessageproperty|Incoming Message Properties
           @since PCC2 2.0.3 */
        return IntMappedMessageValue::create(pl.getAssociatedMessages());
     case ippMinedD:
        /* @q Mined.D:Int (Planet Property)
           Mined Duranium, in kilotons. */
        return makeIntFromMaybe(pl.getCargo(el_Duranium));
     case ippMinedM:
        /* @q Mined.M:Int (Planet Property)
           Mined Molybdenum, in kilotons. */
        return makeIntFromMaybe(pl.getCargo(el_Molybdenum));
     case ippMinedN:
        /* @q Mined.N:Int (Planet Property)
           Mined Neutronium, in kilotons. */
        return makeIntFromMaybe(pl.getCargo(el_Neutronium));
     case ippMinedStr:
        /* @q Mined.Str:Cargo (Planet Property)
           Mined minerals, as a string. */
        {
            GCargoSpec sp;
            bool haveAny = false;
            for (GCargoType el = el_Neutronium; el <= el_Molybdenum; ++el) {
                mn32_t n = pl.getCargo(el);
                if (n.isKnown()) {
                    sp.set(el, n);
                    haveAny = true;
                }
            }
            if (haveAny)
                return makeStringValue(sp.toCargoSpecString());
            else
                return 0;
        }
     case ippMinedT:
        /* @q Mined.T:Int (Planet Property)
           Mined Tritanium, in kilotons. */
        return makeIntFromMaybe(pl.getCargo(el_Tritanium));
     case ippMines:
        /* @q Mines:Int (Planet Property)
           Number of mineral mines. */
        return makeIntFromMaybe(pl.getStructures(str_Mines));
     case ippMinesMax:
        /* @q Mines.Max:Int (Planet Property)
           Maximum number of mineral mines. */
        return makeIntFromMaybe(getMaxStructures(pl, str_Mines));
     case ippMinesWanted:
        /* @q Mines.Want:Int (Planet Property)
           Auto-build goal for mineral mines.
           @assignable */
        pl.getAutobuild(abs);
        return makeIntValue(abs.goal[str_Mines]);
     case ippMoney:
        /* @q Money:Int (Planet Property)
           Money (megacredits) on planet. */
        return makeIntFromMaybe(pl.getCargo(el_Money));
     case ippName:
        /* @q Name:Str (Planet Property)
           Name of planet. */
        return makeStringValue(pl.getName(GObject::PlainName));
     case ippNativeChange:
        /* @q Natives.Change$:Int (Planet Property)
           Native happiness change, numeric value. */
        if (pl.getNatives().isKnown() && pl.getNatives() > 0)
            return makeIntFromMaybe(getNativeChange(pl));
        else
            return 0;
     case ippNativeChangeStr:
        /* @q Natives.Change:Str (Planet Property)
           Native happiness change, text. */
        if (pl.getNatives().isKnown() && pl.getNatives() > 0 && getNativeChange(pl).isKnown())
            return makeStringValue(getHappyChangeName(getNativeChange(pl)));
        else
            return 0;
     case ippNativeGov:
        /* @q Natives.Gov:Str (Planet Property)
           Native government name. */
        if (pl.getNatives().isKnown() && pl.getNatives() > 0 && pl.getNativeGovernment().isKnown())
            return makeStringValue(getNativeGovernmentName(pl.getNativeGovernment()));
        else
            return 0;
     case ippNativeGovCode:
        /* @q Natives.Gov$:Int (Planet Property)
           Native government code. */
        if (pl.getNatives().isKnown() && pl.getNatives() > 0)
            return makeIntFromMaybe(pl.getNativeGovernment());
        else
            return 0;
     case ippNativeHappy:
        /* @q Natives.Happy$:Int (Planet Property)
           Native happiness, numeric value. */
        if (pl.getNatives().isKnown() && pl.getNatives() > 0)
            return makeIntFromMaybe(pl.getNativeHappiness());
        else
            return 0;
     case ippNativeHappyStr:
        /* @q Natives.Happy:Str (Planet Property)
           Native happiness, text. */
        if (pl.getNatives().isKnown() && pl.getNatives() > 0 && pl.getNativeHappiness().isKnown())
            return makeStringValue(getHappinessName(pl.getNativeHappiness()));
        else
            return 0;
     case ippNativeRace:
        /* @q Natives.Race:Str (Planet Property)
           Native race, name. */
        if (pl.getNatives().isKnown() && pl.getNatives() > 0 && pl.getNativeRace().isKnown())
            return makeStringValue(getNativeRaceName(pl.getNativeRace()));
        else
            return 0;
     case ippNativeRaceCode:
        /* @q Natives.Race$:Int (Planet Property)
           Native race, numeric value. */
        if (pl.getNatives().isKnown())
            return makeIntFromMaybe(pl.getNativeRace());
        else
            return 0;
     case ippNativeTax:
        /* @q Natives.Tax:Int (Planet Property)
           Native tax level.
           @assignable
           @see SetNativeTax (Planet Command) */
        if (pl.getNatives().isKnown() && pl.getNatives() > 0)
            return makeIntFromMaybe(pl.getNativeTax());
        else
            return 0;
     case ippNativeTaxBase:
        /* @q Natives.Tax.Base:Int (Planet Property)
           Natives base tax level.
           This is the tax level at which happiness does not change.
           @since PCC2 1.99.15 */
        return makeIntFromMaybe(getNativeBaseTax(pl, 0));
     case ippNativeTaxMax:
        /* @q Natives.Tax.Max:Int (Planet Property)
           Natives maximum tax level.
           This is the tax level at which happiness changes by -30.
           @since PCC2 1.99.15 */
        return makeIntFromMaybe(getNativeBaseTax(pl, -30));
     case ippNativeTaxIncome:
        /* @q Natives.Tax.Income:Int (Planet Property)
           Tax income from natives, megacredits.
           @since PCC2 1.99.15 */
        if (pl.getNatives().isKnown()
            && pl.getNativeRace().isKnown()
            && pl.getNativeGovernment().isKnown()
            && pl.getNativeTax().isKnown())
        {
            if (pl.getNatives() > 0 && pl.isOwnerKnown() && pl.getOwner() > 0) {
                return makeIntValue(getNativeDue(pl, pl.getNativeTax()));
            } else {
                // No natives, or no owner: income is 0
                return makeIntValue(0);
            }
        } else {
            return 0;
        }
     case ippNatives:
        /* @q Natives:Int (Planet Property)
           Native population size, clans. */
        return makeIntFromMaybe(pl.getNatives());
     case ippOrbitingEnemies:
        /* @q Orbit.Enemy:Int (Planet Property)
           Number of enemy (=not own) ships in orbit of this planet. */
        return makeIntValue(countShips(univ, pl.getPos(), GPlayerSet::allUpTo(NUM_OWNERS) - getPlayerId()));
     case ippOrbitingOwn:
        /* @q Orbit.Own:Int (Planet Property)
           Number of own ships in orbit of this planet. */
        return makeIntValue(countShips(univ, pl.getPos(), GPlayerSet(getPlayerId())));
     case ippOrbitingShips:
        /* @q Orbit:Int (Planet Property)
           Total number of ships in orbit of this planet. */
        return makeIntValue(countShips(univ, pl.getPos(), GPlayerSet::allUpTo(NUM_OWNERS)));
     case ippPlayed:
        /* @q Played:Bool (Planet Property)
           True if this planet is played.
           @since PCC 1.1.19 */
        return makeBoolValue(pl.isPlayable(GObject::Playable));
     case ippSupplies:
        /* @q Supplies:Int (Planet Property)
           Supplies on this planet. */
        return makeIntFromMaybe(pl.getCargo(el_Supplies));
     case ippTask:
        /* @q Task:Bool (Planet Property)
           True if this planet has an Auto Task. */
        return makeBoolValue(getProcessByObject(&pl, IntExecutionContext::pkPlanetTask) != 0);
     case ippTaskBase:
        /* @q Task.Base:Bool (Planet Property)
           True if this planet's starbase has an Auto Task. */
        return makeBoolValue(getProcessByObject(&pl, IntExecutionContext::pkBaseTask) != 0);
     case ippTemp:
        /* @q Temp$:Int (Planet Property)
           Temperature, numeric value. */
        return makeIntFromMaybe(pl.getTemperature());
     case ippTempStr:
        /* @q Temp:Int (Planet Property)
           Temperature class, human-readable. */
        if (pl.getTemperature().isKnown())
            return makeStringValue(getTemperatureName(pl.getTemperature()));
        else
            return 0;
     case ippTypeChar:
        /* @q Type.Short:Str (Planet Property)
           Always "P" for planets.
           @see Type.Short (Ship Property), Type.Short (Combat Participant Property)
           @since PCC2 1.99.21, PCC 1.1.20 */
        return makeStringValue("P");
     case ippTypeStr:
        /* @q Type:Str (Planet Property)
           Always "Planet" for planets.
           @see Type (Ship Property), Type (Combat Participant Property)
           @since PCC2 1.99.21, PCC 1.1.20 */
        return makeStringValue("Planet");

     case ippScore:
        return new IntPlanetArrayProperty(pl, ipp);
    }
    return 0;
}

void
setPlanetProperty(GPlanet& pl, IntPlanetProperty ipp, IntValue* value)
{
    // We cannot assign to anything other than auto-build goals on non-played planets
    if (!pl.isPlayable(GObject::Playable) && (ipp != ippMinesWanted && ipp != ippFactoriesWanted && ipp != ippDefenseWanted && ipp != ippBaseDefenseWanted))
        throw IntError::notAssignable();

    int32_t iv;
    switch (ipp) {
     case ippMinesWanted:
        if (checkIntArg(iv, value, 0, 1000))
            pl.setAutobuildGoal(str_Mines, iv);
        break;
     case ippFactoriesWanted:
        if (checkIntArg(iv, value, 0, 1000))
            pl.setAutobuildGoal(str_Factories, iv);
        break;
     case ippDefenseWanted:
        if (checkIntArg(iv, value, 0, 1000))
            pl.setAutobuildGoal(str_Defense, iv);
        break;
     case ippBaseDefenseWanted:
        if (checkIntArg(iv, value, 0, 1000))
            pl.setAutobuildGoal(str_BaseDefense, iv);
        break;
     case ippColonistTax:
        if (checkIntArg(iv, value, 0, 100))
            pl.setColonistTax(mp16_t(iv));
        break;
     case ippFCode:
        {
            string_t sv;
            if (checkStringArg(sv, value))
                pl.setFCode(sv);
            break;
        }
     case ippNativeTax:
        if (checkIntArg(iv, value, 0, 100)) {
            if (pl.getNativeRace() <= 0)
                throw IntError::notAssignable();
            else
                pl.setNativeTax(mp16_t(iv));
        }
        break;
     default:
        throw IntError::notAssignable();
    }
}

void
callPlanetMethod(GPlanet& pl, GUniverse& univ, IntPlanetMethod ipm, IntExecutionContext& exc, IntArgBlock& args)
{
    planet_methods[ipm](pl, univ, exc, args);
}
